from typing import Optional

from dagster._annotations import experimental
from dagster._core.definitions.auto_materialize_rule import AutoMaterializeRule
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.security import non_secure_md5_hash_str

from ..scheduling_condition import SchedulingResult
from ..scheduling_context import SchedulingContext
from .asset_condition import AssetCondition


@experimental
@whitelist_for_serdes
class RuleCondition(AssetCondition):
    """This class represents the condition that a particular AutoMaterializeRule is satisfied."""

    rule: AutoMaterializeRule

    def get_unique_id(self, *, parent_unique_id: Optional[str], index: Optional[str]) -> str:
        # preserves old (bad) behavior of not including the parent_unique_id to avoid inavlidating
        # old serialized information
        parts = [self.rule.__class__.__name__, self.description]
        return non_secure_md5_hash_str("".join(parts).encode())

    @property
    def description(self) -> str:
        return self.rule.description

    def evaluate(self, context: SchedulingContext) -> SchedulingResult:
        context.legacy_context.root_context.daemon_context.logger.debug(
            f"Evaluating rule: {self.rule.to_snapshot()}"
        )
        evaluation_result = self.rule.evaluate_for_asset(context)
        context.legacy_context.root_context.daemon_context.logger.debug(
            f"Rule returned {evaluation_result.true_subset.size} partitions "
            f"({evaluation_result.end_timestamp - evaluation_result.start_timestamp:.2f} seconds)"
        )
        return evaluation_result



import itertools
from typing import Any, NamedTuple, Optional, Sequence

from dagster._core.asset_graph_view.asset_graph_view import (
    AssetGraphView,
    AssetSlice,
    TemporalContext,
)
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_subset import AssetSubset
from dagster._core.definitions.declarative_scheduling.serialized_objects import (
    AssetConditionEvaluation,
    AssetConditionEvaluationState,
    AssetSubsetWithMetadata,
)
from dagster._core.definitions.metadata import MetadataMapping
from dagster._model import DagsterModel
from dagster._utils import utc_datetime_from_timestamp


class AssetSliceWithMetadata(NamedTuple):
    asset_slice: AssetSlice
    metadata: MetadataMapping

    @staticmethod
    def from_asset_subset_with_metadata(
        *, asset_graph_view: AssetGraphView, asset_subset_with_metadata: AssetSubsetWithMetadata
    ) -> Optional["AssetSliceWithMetadata"]:
        metadata = asset_subset_with_metadata.metadata
        if metadata is None:
            return None
        asset_slice = asset_graph_view.get_asset_slice_from_subset(
            asset_subset_with_metadata.subset
        )
        if asset_slice is None:
            return None
        return AssetSliceWithMetadata(asset_slice=asset_slice, metadata=metadata)


class SchedulingEvaluationResultNode(DagsterModel):
    """Represents the results of evaluating a single condition in the broader evaluation tree."""

    asset_key: AssetKey
    condition_unique_id: str

    true_slice: Optional[AssetSlice]
    candidate_slice: Optional[AssetSlice]
    slices_with_metadata: Sequence[AssetSliceWithMetadata]

    extra_state: Any

    @staticmethod
    def nodes_for_evaluation(
        asset_graph_view: AssetGraphView,
        state: AssetConditionEvaluationState,
        condition_evaluation: AssetConditionEvaluation,
    ) -> Sequence["SchedulingEvaluationResultNode"]:
        unique_id = condition_evaluation.condition_snapshot.unique_id
        asset_key = condition_evaluation.asset_key
        # we store AllPartitionsSubset as a sentinel value to avoid serializing the entire set of
        # partitions on each tick. this logic handles converting that back to an AllPartitionsSubset
        # at read time.
        candidate_subset = (
            condition_evaluation.candidate_subset
            if isinstance(condition_evaluation.candidate_subset, AssetSubset)
            else asset_graph_view.get_asset_slice(asset_key).convert_to_valid_asset_subset()
        )

        # only carry forward subsets that are compatible with the current partitions def
        slices_with_metadata = list(
            filter(
                None,
                [
                    AssetSliceWithMetadata.from_asset_subset_with_metadata(
                        asset_graph_view=asset_graph_view,
                        asset_subset_with_metadata=subset_with_metadata,
                    )
                    for subset_with_metadata in condition_evaluation.subsets_with_metadata
                ],
            )
        )
        node = SchedulingEvaluationResultNode(
            asset_key=asset_key,
            condition_unique_id=unique_id,
            true_slice=asset_graph_view.get_asset_slice_from_subset(
                condition_evaluation.true_subset
            ),
            candidate_slice=asset_graph_view.get_asset_slice_from_subset(candidate_subset),
            slices_with_metadata=slices_with_metadata,
            extra_state=state.extra_state_by_unique_id.get(unique_id),
        )
        child_nodes = [
            SchedulingEvaluationResultNode.nodes_for_evaluation(asset_graph_view, state, child)
            for child in condition_evaluation.child_evaluations
        ]
        return list(itertools.chain([node], *child_nodes))


class SchedulingEvaluationInfo(DagsterModel):
    """Represents computed information for the entire evaluation tree."""

    temporal_context: TemporalContext
    evaluation_nodes: Sequence[SchedulingEvaluationResultNode]
    requested_slice: Optional[AssetSlice]

    def get_evaluation_node(self, unique_id: str) -> Optional[SchedulingEvaluationResultNode]:
        for node in self.evaluation_nodes:
            if node.condition_unique_id == unique_id:
                return node
        return None

    @staticmethod
    def from_asset_condition_evaluation_state(
        asset_graph_view: AssetGraphView, state: AssetConditionEvaluationState
    ) -> "SchedulingEvaluationInfo":
        temporal_context = TemporalContext(
            effective_dt=utc_datetime_from_timestamp(state.previous_tick_evaluation_timestamp or 0),
            last_event_id=state.max_storage_id,
        )
        nodes = SchedulingEvaluationResultNode.nodes_for_evaluation(
            asset_graph_view, state, state.previous_evaluation
        )
        requested_slice = asset_graph_view.get_asset_slice_from_subset(state.true_subset)
        return SchedulingEvaluationInfo(
            temporal_context=temporal_context,
            evaluation_nodes=nodes,
            requested_slice=requested_slice,
        )



import datetime
from typing import Sequence, cast

from dagster._core.definitions.declarative_scheduling.scheduling_evaluation_info import (
    AssetSliceWithMetadata,
)
from dagster._core.definitions.declarative_scheduling.utils import SerializableTimeDelta
from dagster._core.definitions.metadata.metadata_value import MetadataValue
from dagster._serdes.serdes import whitelist_for_serdes

from ..scheduling_condition import SchedulingCondition, SchedulingResult
from ..scheduling_context import SchedulingContext

_REQUEST_TIMESTAMP_METADATA_KEY = "request_timestamp"


@whitelist_for_serdes
class ScheduledSinceCondition(SchedulingCondition):
    """SchedulingCondition which is true if the asset has been requested for materialization via
    the declarative scheduling system within the given time window.

    Will only detect requests which have been made since this condition was added to the asset.
    """

    serializable_lookback_timedelta: SerializableTimeDelta

    @property
    def description(self) -> str:
        return f"Has been requested within the last {self.lookback_timedelta}"

    @property
    def lookback_timedelta(self) -> datetime.timedelta:
        return self.serializable_lookback_timedelta.to_timedelta()

    @staticmethod
    def from_lookback_delta(lookback_delta: datetime.timedelta) -> "ScheduledSinceCondition":
        return ScheduledSinceCondition(
            serializable_lookback_timedelta=SerializableTimeDelta.from_timedelta(lookback_delta)
        )

    def _get_minimum_timestamp(self, context: SchedulingContext) -> float:
        """The minimum timestamp for a request to be considered in the lookback window."""
        return (context.effective_dt - self.lookback_timedelta).timestamp()

    def _get_new_slices_with_metadata(
        self, context: SchedulingContext
    ) -> Sequence[AssetSliceWithMetadata]:
        """Updates the stored information as to when the asset was last requested."""
        # the first time this asset has been evaluated
        if context.previous_evaluation_info is None:
            return []

        previous_slices_with_metadata = (
            context.previous_evaluation_node.slices_with_metadata
            if context.previous_evaluation_node
            else []
        )

        # no new updates since previous tick
        if context.previous_requested_slice is None:
            return previous_slices_with_metadata

        # for existing subsets, remove references to newly-requested partitions, as these subsets
        # are meant to represent the most recent time that the asset was requested
        slices_with_metadata = [
            AssetSliceWithMetadata(
                asset_slice.compute_difference(context.previous_requested_slice), metadata
            )
            for asset_slice, metadata in previous_slices_with_metadata
        ]

        # for the newly-requested slice, add a new entry indicating that these partitions were
        # requested on the previous tick
        previous_request_timestamp = (
            context.previous_evaluation_info.temporal_context.effective_dt.timestamp()
        )
        slices_with_metadata.append(
            AssetSliceWithMetadata(
                context.previous_requested_slice,
                {_REQUEST_TIMESTAMP_METADATA_KEY: MetadataValue.float(previous_request_timestamp)},
            )
        )

        # finally, evict any empty subsets from the list, and any subsets with an older timestamp
        return [
            asset_slice_with_metadata
            for asset_slice_with_metadata in slices_with_metadata
            if not (
                asset_slice_with_metadata.asset_slice.is_empty
                or cast(
                    float,
                    asset_slice_with_metadata.metadata.get(
                        _REQUEST_TIMESTAMP_METADATA_KEY, MetadataValue.float(0)
                    ).value,
                )
                < self._get_minimum_timestamp(context)
            )
        ]

    def evaluate(self, context: SchedulingContext) -> SchedulingResult:
        slices_with_metadata = self._get_new_slices_with_metadata(context)

        # we keep track of all slices that have been requested within the lookback window, so we can
        # simply compute the union of all of these slices to determine the true slice
        requested_within_lookback_slice = context.asset_graph_view.create_empty_slice(
            context.asset_key
        )
        for asset_slice, _ in slices_with_metadata:
            requested_within_lookback_slice = requested_within_lookback_slice.compute_union(
                asset_slice
            )

        return SchedulingResult.create(
            context=context,
            true_slice=context.candidate_slice.compute_intersection(
                requested_within_lookback_slice
            ),
            slices_with_metadata=slices_with_metadata,
        )



import glob
import os
from typing import Optional, Sequence, Union

import click

import dagster._check as check
from dagster import (
    DagsterInvariantViolationError,
    __version__ as dagster_version,
)
from dagster._cli.workspace.cli_target import (
    get_external_repository_from_kwargs,
    repository_target_argument,
)
from dagster._core.definitions.run_request import InstigatorType
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation import ExternalRepository
from dagster._core.scheduler.instigation import InstigatorStatus
from dagster._core.scheduler.scheduler import DagsterDaemonScheduler

from .utils import get_instance_for_cli


@click.group(name="schedule")
def schedule_cli():
    """Commands for working with Dagster schedules."""


def print_changes(external_repository, instance, print_fn=print, preview=False):
    debug_info = instance.scheduler_debug_info()
    errors = debug_info.errors
    external_schedules = external_repository.get_external_schedules()
    schedule_states = instance.all_instigator_state(
        external_repository.get_external_origin_id(),
        external_repository.selector_id,
        InstigatorType.SCHEDULE,
    )
    external_schedules_dict = {s.get_external_origin_id(): s for s in external_schedules}
    schedule_states_dict = {s.instigator_origin_id: s for s in schedule_states}

    external_schedule_origin_ids = set(external_schedules_dict.keys())
    schedule_state_ids = set(schedule_states_dict.keys())

    added_schedules = external_schedule_origin_ids - schedule_state_ids
    removed_schedules = schedule_state_ids - external_schedule_origin_ids

    changed_schedules = []
    for schedule_origin_id in external_schedule_origin_ids & schedule_state_ids:
        schedule_state = schedule_states_dict[schedule_origin_id]
        external_schedule = external_schedules_dict[schedule_origin_id]
        if schedule_state.instigator_data.cron_schedule != external_schedule.cron_schedule:
            changed_schedules.append(schedule_origin_id)

    if not errors and not added_schedules and not changed_schedules and not removed_schedules:
        if preview:
            print_fn(click.style("No planned changes to schedules.", fg="magenta", bold=True))
            print_fn(f"{len(external_schedules)} schedules will remain unchanged")
        else:
            print_fn(click.style("No changes to schedules.", fg="magenta", bold=True))
            print_fn(f"{len(external_schedules)} schedules unchanged")
        return

    if errors:
        print_fn(
            click.style(
                "Planned Error Fixes:" if preview else "Errors Resolved:", fg="magenta", bold=True
            )
        )
        print_fn("\n".join(debug_info.errors))

    if added_schedules or changed_schedules or removed_schedules:
        print_fn(
            click.style(
                "Planned Schedule Changes:" if preview else "Changes:", fg="magenta", bold=True
            )
        )

    for schedule_origin_id in added_schedules:
        print_fn(
            click.style(
                f"  + {external_schedules_dict[schedule_origin_id].name} (add) [{schedule_origin_id}]",
                fg="green",
            )
        )

    for schedule_origin_id in changed_schedules:
        schedule_state = schedule_states_dict[schedule_origin_id]
        external_schedule = external_schedules_dict[schedule_origin_id]

        print_fn(
            click.style(
                f"  ~ {external_schedule.name} (update) [{schedule_origin_id}]",
                fg="yellow",
            )
        )
        print_fn(
            click.style("\t cron_schedule: ", fg="yellow")
            + click.style(schedule_state.instigator_data.cron_schedule, fg="red")
            + " => "
            + click.style(external_schedule.cron_schedule, fg="green")
        )

    for schedule_origin_id in removed_schedules:
        print_fn(
            click.style(
                f"  - {schedule_states_dict[schedule_origin_id].instigator_name} (delete) [{schedule_origin_id}]",
                fg="red",
            )
        )


def check_repo_and_scheduler(repository: ExternalRepository, instance: DagsterInstance) -> None:
    check.inst_param(repository, "repository", ExternalRepository)
    check.inst_param(instance, "instance", DagsterInstance)

    repository_name = repository.name

    if not repository.get_external_schedules():
        raise click.UsageError(f"There are no schedules defined for repository {repository_name}.")

    if not os.getenv("DAGSTER_HOME"):
        raise click.UsageError(
            "The environment variable $DAGSTER_HOME is not set. Dagster requires this "
            "environment variable to be set to an existing directory in your filesystem "
            "that contains your dagster instance configuration file (dagster.yaml).\n"
            "You can resolve this error by exporting the environment variable."
            "For example, you can run the following command in your shell or "
            "include it in your shell configuration file:\n"
            '\texport DAGSTER_HOME="~/dagster_home"'
            "\n\n"
        )


@schedule_cli.command(
    name="preview", help="Preview changes that will be performed by `dagster schedule up`."
)
@repository_target_argument
def schedule_preview_command(**kwargs):
    return execute_preview_command(kwargs, click.echo)


def execute_preview_command(cli_args, print_fn):
    with get_instance_for_cli() as instance:
        with get_external_repository_from_kwargs(
            instance, version=dagster_version, kwargs=cli_args
        ) as external_repo:
            check_repo_and_scheduler(external_repo, instance)

            print_changes(external_repo, instance, print_fn, preview=True)


@schedule_cli.command(
    name="list",
    help="List all schedules that correspond to a repository.",
)
@repository_target_argument
@click.option("--running", help="Filter for running schedules", is_flag=True, default=False)
@click.option("--stopped", help="Filter for stopped schedules", is_flag=True, default=False)
@click.option("--name", help="Only display schedule schedule names", is_flag=True, default=False)
def schedule_list_command(running, stopped, name, **kwargs):
    return execute_list_command(running, stopped, name, kwargs, click.echo)


def execute_list_command(running_filter, stopped_filter, name_filter, cli_args, print_fn):
    with get_instance_for_cli() as instance:
        with get_external_repository_from_kwargs(
            instance, version=dagster_version, kwargs=cli_args
        ) as external_repo:
            check_repo_and_scheduler(external_repo, instance)

            repository_name = external_repo.name

            if not name_filter:
                title = f"Repository {repository_name}"
                print_fn(title)
                print_fn("*" * len(title))

            repo_schedules = external_repo.get_external_schedules()
            stored_schedules_by_origin_id = {
                stored_schedule_state.instigator_origin_id: stored_schedule_state
                for stored_schedule_state in instance.all_instigator_state(
                    external_repo.get_external_origin_id(),
                    external_repo.selector_id,
                    instigator_type=InstigatorType.SCHEDULE,
                )
            }

            first = True

            for external_schedule in repo_schedules:
                schedule_state = external_schedule.get_current_instigator_state(
                    stored_schedules_by_origin_id.get(external_schedule.get_external_origin_id())
                )

                if running_filter and not schedule_state.is_running:
                    continue
                if stopped_filter and schedule_state.is_running:
                    continue

                if name_filter:
                    print_fn(external_schedule.name)
                    continue

                status = "RUNNING" if schedule_state.is_running else "STOPPED"
                schedule_title = f"Schedule: {external_schedule.name} [{status}]"
                if not first:
                    print_fn("*" * len(schedule_title))

                first = False

                print_fn(schedule_title)
                print_fn(f"Cron Schedule: {external_schedule.cron_schedule}")


def extract_schedule_name(schedule_name: Optional[Union[str, Sequence[str]]]) -> Optional[str]:
    if schedule_name and not isinstance(schedule_name, str):
        if len(schedule_name) == 1:
            return schedule_name[0]
        else:
            check.failed(f"Can only handle zero or one schedule args. Got {schedule_name!r}")
    return None


@schedule_cli.command(name="start", help="Start an existing schedule.")
@click.argument("schedule_name", nargs=-1)  # , required=True)
@click.option("--start-all", help="start all schedules", is_flag=True, default=False)
@repository_target_argument
def schedule_start_command(schedule_name, start_all, **kwargs):
    schedule_name = extract_schedule_name(schedule_name)
    if schedule_name is None and start_all is False:
        print(  # noqa: T201
            "Noop: dagster schedule start was called without any arguments specifying which "
            "schedules to start. Pass a schedule name or the --start-all flag to start schedules."
        )
        return
    return execute_start_command(schedule_name, start_all, kwargs, click.echo)


def execute_start_command(schedule_name, all_flag, cli_args, print_fn):
    with get_instance_for_cli() as instance:
        with get_external_repository_from_kwargs(
            instance, version=dagster_version, kwargs=cli_args
        ) as external_repo:
            check_repo_and_scheduler(external_repo, instance)

            repository_name = external_repo.name

            if all_flag:
                for external_schedule in external_repo.get_external_schedules():
                    try:
                        instance.start_schedule(external_schedule)
                    except DagsterInvariantViolationError as ex:
                        raise click.UsageError(ex)

                print_fn(f"Started all schedules for repository {repository_name}")
            else:
                try:
                    instance.start_schedule(external_repo.get_external_schedule(schedule_name))
                except DagsterInvariantViolationError as ex:
                    raise click.UsageError(ex)

                print_fn(f"Started schedule {schedule_name}")


@schedule_cli.command(name="stop", help="Stop an existing schedule.")
@click.argument("schedule_name", nargs=-1)
@repository_target_argument
def schedule_stop_command(schedule_name, **kwargs):
    schedule_name = extract_schedule_name(schedule_name)
    return execute_stop_command(schedule_name, kwargs, click.echo)


def execute_stop_command(schedule_name, cli_args, print_fn, instance=None):
    with get_instance_for_cli() as instance:
        with get_external_repository_from_kwargs(
            instance, version=dagster_version, kwargs=cli_args
        ) as external_repo:
            check_repo_and_scheduler(external_repo, instance)

            try:
                external_schedule = external_repo.get_external_schedule(schedule_name)
                instance.stop_schedule(
                    external_schedule.get_external_origin_id(),
                    external_schedule.selector_id,
                    external_schedule,
                )
            except DagsterInvariantViolationError as ex:
                raise click.UsageError(ex)

            print_fn(f"Stopped schedule {schedule_name}")


@schedule_cli.command(name="logs", help="Get logs for a schedule.")
@click.argument("schedule_name", nargs=-1)
@repository_target_argument
def schedule_logs_command(schedule_name, **kwargs):
    schedule_name = extract_schedule_name(schedule_name)
    if schedule_name is None:
        print(  # noqa: T201
            "Noop: dagster schedule logs was called without any arguments specifying which "
            "schedules to retrieve logs for. Pass a schedule name"
        )
        return
    return execute_logs_command(schedule_name, kwargs, click.echo)


def execute_logs_command(schedule_name, cli_args, print_fn, instance=None):
    with get_instance_for_cli() as instance:
        with get_external_repository_from_kwargs(
            instance, version=dagster_version, kwargs=cli_args
        ) as external_repo:
            check_repo_and_scheduler(external_repo, instance)

            if isinstance(instance.scheduler, DagsterDaemonScheduler):
                return print_fn(
                    "This command is deprecated for the DagsterDaemonScheduler. "
                    "Logs for the DagsterDaemonScheduler written to the process output. "
                    "For help troubleshooting the Daemon Scheduler, see "
                    "https://docs.dagster.io/troubleshooting/schedules"
                )

            logs_path = os.path.join(
                instance.logs_path_for_schedule(
                    external_repo.get_external_schedule(schedule_name).get_external_origin_id()
                )
            )

            logs_directory = os.path.dirname(logs_path)
            result_files = glob.glob(f"{logs_directory}/*.result")
            most_recent_log = max(result_files, key=os.path.getctime) if result_files else None

            output = ""

            title = "Scheduler Logs:"
            output += "{title}\n{sep}\n{info}\n".format(
                title=title,
                sep="=" * len(title),
                info=logs_path,
            )

            title = (
                "Schedule Execution Logs:"
                "\nEvent logs from schedule execution. "
                "Errors that caused schedule executions to not run or fail can be found here. "
            )
            most_recent_info = (
                f"\nMost recent execution log: {most_recent_log}" if most_recent_log else ""
            )
            info = f"All execution logs: {logs_directory}{most_recent_info}"
            output += "\n{title}\n{sep}\n{info}\n".format(
                title=title,
                sep="=" * len(title),
                info=info,
            )

            print_fn(output)


@schedule_cli.command(name="restart", help="Restart a running schedule.")
@click.argument("schedule_name", nargs=-1)
@click.option(
    "--restart-all-running",
    help="restart previously running schedules",
    is_flag=True,
    default=False,
)
@repository_target_argument
def schedule_restart_command(schedule_name, restart_all_running, **kwargs):
    schedule_name = extract_schedule_name(schedule_name)
    return execute_restart_command(schedule_name, restart_all_running, kwargs, click.echo)


def execute_restart_command(schedule_name, all_running_flag, cli_args, print_fn):
    with get_instance_for_cli() as instance:
        with get_external_repository_from_kwargs(
            instance, version=dagster_version, kwargs=cli_args
        ) as external_repo:
            check_repo_and_scheduler(external_repo, instance)

            repository_name = external_repo.name

            if all_running_flag:
                for schedule_state in instance.all_instigator_state(
                    external_repo.get_external_origin_id(),
                    external_repo.selector_id,
                    InstigatorType.SCHEDULE,
                ):
                    if schedule_state.status == InstigatorStatus.RUNNING:
                        try:
                            external_schedule = external_repo.get_external_schedule(
                                schedule_state.instigator_name
                            )
                            instance.stop_schedule(
                                schedule_state.instigator_origin_id,
                                external_schedule.selector_id,
                                external_schedule,
                            )
                            instance.start_schedule(external_schedule)
                        except DagsterInvariantViolationError as ex:
                            raise click.UsageError(ex)

                print_fn(f"Restarted all running schedules for repository {repository_name}")
            else:
                external_schedule = external_repo.get_external_schedule(schedule_name)
                schedule_state = instance.get_instigator_state(
                    external_schedule.get_external_origin_id(),
                    external_schedule.selector_id,
                )
                if schedule_state is not None and schedule_state.status != InstigatorStatus.RUNNING:
                    click.UsageError(
                        f"Cannot restart a schedule {schedule_state.instigator_name} because is not currently running"
                    )

                try:
                    instance.stop_schedule(
                        schedule_state.instigator_origin_id,
                        external_schedule.selector_id,
                        external_schedule,
                    )
                    instance.start_schedule(external_schedule)
                except DagsterInvariantViolationError as ex:
                    raise click.UsageError(ex)

                print_fn(f"Restarted schedule {schedule_name}")


@schedule_cli.command(name="wipe", help="Delete the schedule history and turn off all schedules.")
def schedule_wipe_command():
    return execute_wipe_command(click.echo)


def execute_wipe_command(print_fn):
    with get_instance_for_cli() as instance:
        confirmation = click.prompt(
            "Are you sure you want to turn off all schedules and delete all schedule history? Type"
            " DELETE"
        )
        if confirmation == "DELETE":
            instance.wipe_all_schedules()
            print_fn("Turned off all schedules and deleted all schedule history")
        else:
            print_fn("Exiting without turning off schedules or deleting schedule history")


@schedule_cli.command(name="debug", help="Debug information about the scheduler.")
def schedule_debug_command():
    return execute_debug_command(click.echo)


def execute_debug_command(print_fn):
    with get_instance_for_cli() as instance:
        debug_info = instance.scheduler_debug_info()

        output = ""

        errors = debug_info.errors
        if len(errors):
            title = "Errors (Run `dagster schedule up` to resolve)"
            output += "\n{title}\n{sep}\n{info}\n\n".format(
                title=title,
                sep="=" * len(title),
                info="\n".join(debug_info.errors),
            )

        title = "Scheduler Configuration"
        output += "{title}\n{sep}\n{info}\n".format(
            title=title,
            sep="=" * len(title),
            info=debug_info.scheduler_config_info,
        )

        title = "Scheduler Info"
        output += "{title}\n{sep}\n{info}\n".format(
            title=title, sep="=" * len(title), info=debug_info.scheduler_info
        )

        title = "Scheduler Storage Info"
        output += "\n{title}\n{sep}\n{info}\n".format(
            title=title,
            sep="=" * len(title),
            info="\n".join(debug_info.schedule_storage),
        )

        print_fn(output)



from typing import AbstractSet

from dagster._core.asset_graph_view.asset_graph_view import AssetSlice
from dagster._core.definitions.declarative_scheduling.scheduling_condition import (
    SchedulingCondition,
    SchedulingResult,
)
from dagster._core.definitions.declarative_scheduling.scheduling_context import SchedulingContext
from dagster._core.definitions.events import AssetKeyPartitionKey
from dagster._serdes.serdes import whitelist_for_serdes


@whitelist_for_serdes
class ParentNewerCondition(SchedulingCondition):
    @property
    def description(self) -> str:
        return "At least one parent has been updated more recently than the candidate."

    def compute_newer_parents(
        self, context: SchedulingContext, candidate: AssetKeyPartitionKey
    ) -> AbstractSet[AssetKeyPartitionKey]:
        """Returns the set of parent asset partitions that are newer than the candidate."""
        # TODO: replace this implementation with one backed by the StaleStatusResolver
        candidate_slice = context.asset_graph_view.get_asset_slice_from_asset_partitions(
            {candidate}
        )
        parent_slices = candidate_slice.compute_parent_slices()
        all_parent_asset_partitions = {
            ap
            for parent_slice in parent_slices.values()
            for ap in parent_slice.expensively_compute_asset_partitions()
        }
        return (
            context.legacy_context.instance_queryer.get_parent_asset_partitions_updated_after_child(
                asset_partition=candidate,
                parent_asset_partitions=all_parent_asset_partitions,
                respect_materialization_data_versions=False,
                ignored_parent_keys=set(),
            )
        )

    def compute_parent_newer_slice(
        self, context: SchedulingContext, from_slice: AssetSlice
    ) -> AssetSlice:
        """For a given slice, computes the slice for which at least one parent is newer than the
        given asset partition.
        """
        asset_partitions_with_updated_parents = {
            ap
            for ap in from_slice.expensively_compute_asset_partitions()
            if len(self.compute_newer_parents(context, ap)) > 0
        }
        return (
            context.asset_graph_view.get_asset_slice_from_asset_partitions(
                asset_partitions_with_updated_parents
            )
            if asset_partitions_with_updated_parents
            else context.asset_graph_view.create_empty_slice(context.asset_key)
        )

    def compute_slice_to_evaluate(self, context: SchedulingContext) -> AssetSlice:
        """Computes the slice of the target asset that must be fully evaluated this tick."""
        if context.previous_evaluation_max_storage_id is None:
            # on the first tick, evaluate all candidates, as you have no previous information
            return context.candidate_slice
        else:
            # on subsequent ticks, only evaluate candidates if they were not evaluated on the
            # previous tick, or have been updated since the previous tick, or have parents that
            # have been updated since the previous tick
            if context.previous_candidate_slice:
                new_candidates = context.candidate_slice.compute_difference(
                    context.previous_candidate_slice
                )
            else:
                new_candidates = context.candidate_slice

            # candidates that have updated since the previous tick
            newly_updated_slice = context.asset_graph_view.compute_updated_since_cursor_slice(
                asset_key=context.asset_key, cursor=context.previous_evaluation_max_storage_id
            )
            # candidates that have parents that have updated since the previous tick
            parent_newly_updated_slice = (
                context.asset_graph_view.compute_parent_updated_since_cursor_slice(
                    asset_key=context.asset_key,
                    cursor=context.previous_evaluation_max_storage_id,
                )
            )
            updated_slice = newly_updated_slice.compute_union(parent_newly_updated_slice)

            return new_candidates.compute_union(
                context.candidate_slice.compute_intersection(updated_slice)
            )

    def evaluate(self, context: SchedulingContext):
        slice_to_evaluate = self.compute_slice_to_evaluate(context)
        new_parent_newer_slice = self.compute_parent_newer_slice(context, slice_to_evaluate)

        if context.previous_evaluation_node and context.previous_evaluation_node.true_slice:
            # combine new results calculated this tick with results from the previous tick
            true_slice = context.previous_evaluation_node.true_slice.compute_difference(
                slice_to_evaluate
            ).compute_union(new_parent_newer_slice)
        else:
            true_slice = new_parent_newer_slice

        return SchedulingResult.create(context, true_slice=true_slice)



from typing import Dict, Mapping, Sequence, Tuple, Union, cast

from pydantic import BaseModel

from dagster._annotations import experimental
from dagster._core.instance import DagsterInstance

from ..asset_check_result import AssetCheckResult
from ..asset_check_spec import AssetCheckKey, AssetCheckSeverity, AssetCheckSpec
from ..asset_checks import AssetChecksDefinition
from ..asset_key import AssetKey, CoercibleToAssetKey
from ..assets import AssetsDefinition, SourceAsset
from ..decorators.asset_check_decorator import multi_asset_check
from ..events import AssetMaterialization
from ..metadata import TableColumn, TableMetadataSet, TableSchema
from .utils import unique_id_from_asset_keys


@experimental
def build_column_schema_change_checks(
    *,
    assets: Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]],
    severity: AssetCheckSeverity = AssetCheckSeverity.WARN,
) -> Sequence[AssetChecksDefinition]:
    """Returns asset checks that pass if the column schema of the asset's latest materialization
    is the same as the column schema of the asset's previous materialization.

    Args:
        assets (Sequence[Union[AssetKey, str, AssetsDefinition, SourceAsset]]): The assets to create
            asset checks for.
        severity (AssetCheckSeverity): The severity if the check fails. Defaults to WARN.

    Returns:
        Sequence[AssetsChecksDefinition]
    """
    asset_keys = set()
    for el in assets:
        if isinstance(el, AssetsDefinition):
            asset_keys |= el.keys
        elif isinstance(el, SourceAsset):
            asset_keys.add(el.key)
        else:
            asset_keys.add(AssetKey.from_coercible(el))

    def _result_for_check_key(
        instance: DagsterInstance, asset_check_key: AssetCheckKey
    ) -> Tuple[bool, str]:
        materialization_records = instance.fetch_materializations(
            limit=2, records_filter=asset_check_key.asset_key
        ).records
        if len(materialization_records) < 2:
            return True, "The asset has been materialized fewer than 2 times"

        record, prev_record = materialization_records
        metadata = cast(AssetMaterialization, record.asset_materialization).metadata
        prev_metadata = cast(AssetMaterialization, prev_record.asset_materialization).metadata

        column_schema = TableMetadataSet.extract(metadata).column_schema
        prev_column_schema = TableMetadataSet.extract(prev_metadata).column_schema
        if column_schema is None:
            return False, "Latest materialization has no column schema metadata"
        if prev_column_schema is None:
            return False, "Previous materialization has no column schema metadata"

        diff = TableSchemaDiff.from_table_schemas(prev_column_schema, column_schema)
        if diff.added_columns or diff.removed_columns or diff.column_type_changes:
            description = "Column schema changed between previous and latest materialization."
            if diff.added_columns:
                description += (
                    f"\n\nAdded columns: {', '.join(col.name for col in diff.added_columns)}"
                )

            if diff.removed_columns:
                description += (
                    f"\n\nRemoved columns: {', '.join(col.name for col in diff.removed_columns)}"
                )

            if diff.column_type_changes:
                description += "\n\nColumn type changes:"
                for col_name, type_change in diff.column_type_changes.items():
                    description += (
                        f"\n- {col_name}: {type_change.old_type} -> {type_change.new_type}"
                    )

            return False, description

        return True, "No changes to column schema between previous and latest materialization"

    @multi_asset_check(
        specs=[
            AssetCheckSpec(
                "column_schema_change",
                asset=asset_key,
                description="Checks whether there are changes to column schema between the asset's "
                " two most recent materializations",
            )
            for asset_key in asset_keys
        ],
        can_subset=True,
        name=unique_id_from_asset_keys(list(asset_keys)),
    )
    def _checks(context):
        for asset_check_key in context.selected_asset_check_keys:
            passed, description = _result_for_check_key(context.instance, asset_check_key)
            yield AssetCheckResult(
                passed=passed,
                description=description,
                severity=severity,
                check_name=asset_check_key.name,
                asset_key=asset_check_key.asset_key,
            )

    return [_checks]


class TypeChange(BaseModel):
    old_type: str
    new_type: str


class TableSchemaDiff(BaseModel):
    added_columns: Sequence[TableColumn]
    removed_columns: Sequence[TableColumn]
    column_type_changes: Mapping[str, TypeChange]

    @staticmethod
    def from_table_schemas(
        old_table_schema: TableSchema, new_table_schema: TableSchema
    ) -> "TableSchemaDiff":
        old_columns_by_key = {column.name: column for column in old_table_schema.columns}
        new_columns_by_key = {column.name: column for column in new_table_schema.columns}

        added_columns = [
            column for column in new_table_schema.columns if column.name not in old_columns_by_key
        ]
        removed_columns = [
            column for column in old_table_schema.columns if column.name not in new_columns_by_key
        ]
        column_type_changes: Dict[str, TypeChange] = {}
        for name, new_column in new_columns_by_key.items():
            old_column = old_columns_by_key.get(name)
            if old_column is not None and old_column.type != new_column.type:
                column_type_changes[name] = TypeChange(
                    old_type=old_column.type, new_type=new_column.type
                )

        return TableSchemaDiff(
            added_columns=added_columns,
            removed_columns=removed_columns,
            column_type_changes=column_type_changes,
        )



import re
from typing import TYPE_CHECKING, Mapping, NamedTuple, Optional, Sequence, Union

import dagster._check as check
import dagster._seven as seven
from dagster._annotations import PublicAttr
from dagster._serdes import whitelist_for_serdes

ASSET_KEY_SPLIT_REGEX = re.compile("[^a-zA-Z0-9_]")
ASSET_KEY_DELIMITER = "/"

if TYPE_CHECKING:
    from dagster._core.definitions.assets import AssetsDefinition
    from dagster._core.definitions.source_asset import SourceAsset


def parse_asset_key_string(s: str) -> Sequence[str]:
    return list(filter(lambda x: x, re.split(ASSET_KEY_SPLIT_REGEX, s)))


@whitelist_for_serdes
class AssetKey(NamedTuple("_AssetKey", [("path", PublicAttr[Sequence[str]])])):
    """Object representing the structure of an asset key.  Takes in a sanitized string, list of
    strings, or tuple of strings.

    Example usage:

    .. code-block:: python

        from dagster import AssetKey

        AssetKey("asset1")
        AssetKey(["asset1"]) # same as the above
        AssetKey(["prefix", "asset1"])
        AssetKey(["prefix", "subprefix", "asset1"])

    Args:
        path (Union[str, Sequence[str]]): String, list of strings, or tuple of strings.  A list of
            strings represent the hierarchical structure of the asset_key.
    """

    def __new__(cls, path: Union[str, Sequence[str]]):
        if isinstance(path, str):
            path = [path]
        else:
            path = list(check.sequence_param(path, "path", of_type=str))

        return super(AssetKey, cls).__new__(cls, path=path)

    def __str__(self):
        return f"AssetKey({self.path})"

    def __repr__(self):
        return f"AssetKey({self.path})"

    def __hash__(self):
        return hash(tuple(self.path))

    def __eq__(self, other):
        if other.__class__ is not self.__class__:
            return False

        return self.path == other.path

    def to_string(self) -> str:
        """E.g. '["first_component", "second_component"]'."""
        return seven.json.dumps(self.path)

    def to_user_string(self) -> str:
        """E.g. "first_component/second_component"."""
        return ASSET_KEY_DELIMITER.join(self.path)

    def to_python_identifier(self, suffix: Optional[str] = None) -> str:
        """Build a valid Python identifier based on the asset key that can be used for
        operation names or I/O manager keys.
        """
        path = list(self.path)

        if suffix is not None:
            path.append(suffix)

        return "__".join(path).replace("-", "_")

    @staticmethod
    def from_user_string(asset_key_string: str) -> "AssetKey":
        return AssetKey(asset_key_string.split(ASSET_KEY_DELIMITER))

    @staticmethod
    def from_db_string(asset_key_string: Optional[str]) -> Optional["AssetKey"]:
        if not asset_key_string:
            return None
        if asset_key_string[0] == "[":
            # is a json string
            try:
                path = seven.json.loads(asset_key_string)
            except seven.JSONDecodeError:
                path = parse_asset_key_string(asset_key_string)
        else:
            path = parse_asset_key_string(asset_key_string)
        return AssetKey(path)

    @staticmethod
    def get_db_prefix(path: Sequence[str]):
        check.sequence_param(path, "path", of_type=str)
        return seven.json.dumps(path)[:-2]  # strip trailing '"]' from json string

    @staticmethod
    def from_graphql_input(graphql_input_asset_key: Mapping[str, Sequence[str]]) -> "AssetKey":
        return AssetKey(graphql_input_asset_key["path"])

    def to_graphql_input(self) -> Mapping[str, Sequence[str]]:
        return {"path": self.path}

    @staticmethod
    def from_coercible(arg: "CoercibleToAssetKey") -> "AssetKey":
        if isinstance(arg, AssetKey):
            return check.inst_param(arg, "arg", AssetKey)
        elif isinstance(arg, str):
            return AssetKey([arg])
        elif isinstance(arg, list):
            check.list_param(arg, "arg", of_type=str)
            return AssetKey(arg)
        elif isinstance(arg, tuple):
            check.tuple_param(arg, "arg", of_type=str)
            return AssetKey(arg)
        else:
            check.failed(f"Unexpected type for AssetKey: {type(arg)}")

    @staticmethod
    def from_coercible_or_definition(
        arg: Union["CoercibleToAssetKey", "AssetsDefinition", "SourceAsset"],
    ) -> "AssetKey":
        from dagster._core.definitions.assets import AssetsDefinition
        from dagster._core.definitions.source_asset import SourceAsset

        if isinstance(arg, AssetsDefinition):
            return arg.key
        elif isinstance(arg, SourceAsset):
            return arg.key
        else:
            return AssetKey.from_coercible(arg)

    def has_prefix(self, prefix: Sequence[str]) -> bool:
        return len(self.path) >= len(prefix) and self.path[: len(prefix)] == prefix

    def with_prefix(self, prefix: "CoercibleToAssetKeyPrefix") -> "AssetKey":
        prefix = key_prefix_from_coercible(prefix)
        return AssetKey(list(prefix) + list(self.path))


CoercibleToAssetKey = Union[AssetKey, str, Sequence[str]]
CoercibleToAssetKeyPrefix = Union[str, Sequence[str]]


def check_opt_coercible_to_asset_key_prefix_param(
    prefix: Optional[CoercibleToAssetKeyPrefix], param_name: str
) -> Optional[Sequence[str]]:
    try:
        return key_prefix_from_coercible(prefix) if prefix is not None else None
    except check.CheckError:
        raise check.ParameterCheckError(
            f'Param "{param_name}" is not a string or a sequence of strings'
        )


def key_prefix_from_coercible(key_prefix: CoercibleToAssetKeyPrefix) -> Sequence[str]:
    if isinstance(key_prefix, str):
        return [key_prefix]
    elif isinstance(key_prefix, list):
        return key_prefix
    else:
        check.failed(f"Unexpected type for key_prefix: {type(key_prefix)}")



from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Dict,
    FrozenSet,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from dagster._core.definitions.asset_subset import AssetSubset, ValidAssetSubset
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.metadata import MetadataMapping, MetadataValue
from dagster._core.definitions.partition import AllPartitionsSubset
from dagster._model import DagsterModel
from dagster._serdes.serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.definitions.declarative_scheduling.scheduling_condition import (
        SchedulingResult,
    )
    from dagster._core.definitions.declarative_scheduling.scheduling_context import (
        SchedulingContext,
    )

T = TypeVar("T")


@whitelist_for_serdes
class HistoricalAllPartitionsSubsetSentinel(DagsterModel):
    """Serializable indicator that this value was an AllPartitionsSubset at serialization time, but
    the partitions may have changed since that time.
    """


def get_serializable_candidate_subset(
    candidate_subset: Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel],
) -> Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel]:
    """Do not serialize the candidate subset directly if it is an AllPartitionsSubset."""
    if isinstance(candidate_subset, AssetSubset) and isinstance(
        candidate_subset.value, AllPartitionsSubset
    ):
        return HistoricalAllPartitionsSubsetSentinel()
    return candidate_subset


@whitelist_for_serdes
class AssetConditionSnapshot(DagsterModel):
    """A serializable snapshot of a node in the AutomationCondition tree."""

    class_name: str
    description: str
    unique_id: str


@whitelist_for_serdes
class AssetSubsetWithMetadata(DagsterModel):
    """An asset subset with metadata that corresponds to it."""

    subset: AssetSubset
    metadata: MetadataMapping

    @property
    def frozen_metadata(self) -> FrozenSet[Tuple[str, MetadataValue]]:
        return frozenset(self.metadata.items())


@whitelist_for_serdes
class AssetConditionEvaluation(DagsterModel):
    """Serializable representation of the results of evaluating a node in the evaluation tree."""

    condition_snapshot: AssetConditionSnapshot
    start_timestamp: Optional[float]
    end_timestamp: Optional[float]

    true_subset: AssetSubset
    candidate_subset: Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel]
    subsets_with_metadata: Sequence[AssetSubsetWithMetadata]

    child_evaluations: Sequence["AssetConditionEvaluation"]

    @property
    def asset_key(self) -> AssetKey:
        return self.true_subset.asset_key

    @staticmethod
    def from_result(result: "SchedulingResult") -> "AssetConditionEvaluation":
        return AssetConditionEvaluation(
            condition_snapshot=result.condition.get_snapshot(result.condition_unique_id),
            start_timestamp=result.start_timestamp,
            end_timestamp=result.end_timestamp,
            true_subset=result.true_subset,
            candidate_subset=get_serializable_candidate_subset(result.candidate_subset),
            subsets_with_metadata=result.subsets_with_metadata,
            child_evaluations=[
                AssetConditionEvaluation.from_result(child_result)
                for child_result in result.child_results
            ],
        )

    def equivalent_to_stored_evaluation(self, other: Optional["AssetConditionEvaluation"]) -> bool:
        """Returns if this evaluation is functionally equivalent to the given stored evaluation.
        This is used to determine if it is necessary to store this new evaluation in the database.
        Noteably, the timestamps on successive evaluations will always be different, so a simple
        equality check would be too strict.
        """
        return (
            other is not None
            and self.condition_snapshot == other.condition_snapshot
            and self.true_subset == other.true_subset
            # the candidate subset gets modified during serialization
            and get_serializable_candidate_subset(self.candidate_subset)
            == get_serializable_candidate_subset(other.candidate_subset)
            and self.subsets_with_metadata == other.subsets_with_metadata
            and len(self.child_evaluations) == len(other.child_evaluations)
            and all(
                self_child.equivalent_to_stored_evaluation(other_child)
                for self_child, other_child in zip(self.child_evaluations, other.child_evaluations)
            )
        )

    def discarded_subset(self) -> Optional[AssetSubset]:
        """Returns the AssetSubset representing asset partitions that were discarded during this
        evaluation. Note that 'discarding' is a deprecated concept that is only used for backwards
        compatibility.
        """
        if len(self.child_evaluations) != 3:
            return None
        not_discard_evaluation = self.child_evaluations[-1]
        discard_evaluation = not_discard_evaluation.child_evaluations[0]
        return discard_evaluation.true_subset

    def get_requested_or_discarded_subset(self) -> AssetSubset:
        discarded_subset = self.discarded_subset()
        if discarded_subset is None:
            return self.true_subset
        else:
            # the true_subset and discarded_subset were created on the same tick, so they are
            # guaranteed to be compatible with each other
            return (
                ValidAssetSubset(asset_key=self.asset_key, value=self.true_subset.value)
                | discarded_subset
            )

    def for_child(self, child_unique_id: str) -> Optional["AssetConditionEvaluation"]:
        """Returns the evaluation of a given child condition by finding the child evaluation that
        has an identical hash to the given condition.
        """
        for child_evaluation in self.child_evaluations:
            if child_evaluation.condition_snapshot.unique_id == child_unique_id:
                return child_evaluation

        return None

    def with_run_ids(self, run_ids: AbstractSet[str]) -> "AssetConditionEvaluationWithRunIds":
        return AssetConditionEvaluationWithRunIds(evaluation=self, run_ids=frozenset(run_ids))

    def legacy_num_skipped(self) -> int:
        if len(self.child_evaluations) < 2:
            return 0

        not_skip_evaluation = self.child_evaluations[-1]
        skip_evaluation = not_skip_evaluation.child_evaluations[0]
        return skip_evaluation.true_subset.size - self.legacy_num_discarded()

    def legacy_num_discarded(self) -> int:
        discarded_subset = self.discarded_subset()
        if discarded_subset is None:
            return 0
        return discarded_subset.size


@whitelist_for_serdes
class AssetConditionEvaluationWithRunIds(DagsterModel):
    """A union of an AssetConditionEvaluation and the set of run IDs that have been launched in
    response to it.
    """

    evaluation: AssetConditionEvaluation
    run_ids: FrozenSet[str]

    @property
    def asset_key(self) -> AssetKey:
        return self.evaluation.asset_key

    @property
    def num_requested(self) -> int:
        return self.evaluation.true_subset.size


@whitelist_for_serdes
@dataclass(frozen=True)
class AssetConditionEvaluationState:
    """Incremental state calculated during the evaluation of an AssetCondition. This may be used
    on the subsequent evaluation to make the computation more efficient.

    Attributes:
        previous_evaluation: The computed AssetConditionEvaluation.
        previous_tick_evaluation_timestamp: The evaluation_timestamp at which the evaluation was performed.
        max_storage_id: The maximum storage ID over all events used in this computation.
        extra_state_by_unique_id: A mapping from the unique ID of each condition in the evaluation
            tree to the extra state that was calculated for it, if any.
    """

    previous_evaluation: AssetConditionEvaluation
    previous_tick_evaluation_timestamp: Optional[float]

    max_storage_id: Optional[int]
    extra_state_by_unique_id: Mapping[str, Optional[Union[AssetSubset, Sequence[AssetSubset]]]]

    @property
    def asset_key(self) -> AssetKey:
        return self.previous_evaluation.asset_key

    @property
    def true_subset(self) -> AssetSubset:
        return self.previous_evaluation.true_subset

    @staticmethod
    def create(
        context: "SchedulingContext", root_result: "SchedulingResult"
    ) -> "AssetConditionEvaluationState":
        """Convenience constructor to generate an AssetConditionEvaluationState from the root result
        and the context in which it was evaluated.
        """

        # flatten the extra state into a single dict
        def _flatten_extra_state(
            r: "SchedulingResult",
        ) -> Mapping[str, Optional[Union[AssetSubset, Sequence[AssetSubset]]]]:
            extra_state: Dict[str, Optional[Union[AssetSubset, Sequence[AssetSubset]]]] = (
                {r.condition_unique_id: r.extra_state} if r.extra_state else {}
            )
            for child in r.child_results:
                extra_state.update(_flatten_extra_state(child))
            return extra_state

        return AssetConditionEvaluationState(
            previous_evaluation=AssetConditionEvaluation.from_result(root_result),
            previous_tick_evaluation_timestamp=context.asset_graph_view.effective_dt.timestamp(),
            max_storage_id=context.new_max_storage_id,
            extra_state_by_unique_id=_flatten_extra_state(root_result),
        )

    def get_extra_state(self, unique_id: str, as_type: Type[T]) -> Optional[T]:
        """Returns the value from the extras dict for the given condition, if it exists and is of
        the expected type. Otherwise, returns None.
        """
        extra_state = self.extra_state_by_unique_id.get(unique_id)
        if isinstance(extra_state, as_type):
            return extra_state
        return None



import logging
import sys
import threading
from collections import defaultdict
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import AbstractContextManager
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)

import pendulum
from typing_extensions import Self

import dagster._check as check
import dagster._seven as seven
from dagster._core.definitions.run_request import (
    AddDynamicPartitionsRequest,
    DagsterRunReaction,
    DeleteDynamicPartitionsRequest,
    InstigatorType,
    RunRequest,
)
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.definitions.sensor_definition import (
    DefaultSensorStatus,
    SensorType,
)
from dagster._core.definitions.utils import normalize_tags
from dagster._core.errors import DagsterError
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation.code_location import CodeLocation
from dagster._core.remote_representation.external import ExternalJob, ExternalSensor
from dagster._core.remote_representation.external_data import ExternalTargetData
from dagster._core.scheduler.instigation import (
    DynamicPartitionsRequestResult,
    InstigatorState,
    InstigatorStatus,
    InstigatorTick,
    SensorInstigatorData,
    TickData,
    TickStatus,
)
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus, RunsFilter
from dagster._core.storage.tags import RUN_KEY_TAG, SENSOR_NAME_TAG
from dagster._core.telemetry import SENSOR_RUN_CREATED, hash_name, log_action
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._daemon.utils import DaemonErrorCapture
from dagster._scheduler.stale import resolve_stale_or_missing_assets
from dagster._utils import DebugCrashFlags, SingleInstigatorDebugCrashFlags, check_for_debug_crash
from dagster._utils.error import SerializableErrorInfo
from dagster._utils.merger import merge_dicts

if TYPE_CHECKING:
    from pendulum.datetime import DateTime

    from dagster._daemon.daemon import DaemonIterator


MIN_INTERVAL_LOOP_TIME = 5

FINISHED_TICK_STATES = [TickStatus.SKIPPED, TickStatus.SUCCESS, TickStatus.FAILURE]


class DagsterSensorDaemonError(DagsterError):
    """Error when running the SensorDaemon."""


class SkippedSensorRun(NamedTuple):
    """Placeholder for runs that are skipped during the run_key idempotence check."""

    run_key: Optional[str]
    existing_run: DagsterRun


class SensorLaunchContext(AbstractContextManager):
    def __init__(
        self,
        external_sensor: ExternalSensor,
        tick: InstigatorTick,
        instance: DagsterInstance,
        logger: logging.Logger,
        tick_retention_settings,
    ):
        self._external_sensor = external_sensor
        self._instance = instance
        self._logger = logger
        self._tick = tick
        self._should_update_cursor_on_failure = False
        self._purge_settings = defaultdict(set)
        for status, day_offset in tick_retention_settings.items():
            self._purge_settings[day_offset].add(status)

    @property
    def status(self) -> TickStatus:
        return self._tick.status

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @property
    def run_count(self) -> int:
        return len(self._tick.run_ids)

    @property
    def tick_id(self) -> str:
        return str(self._tick.tick_id)

    @property
    def log_key(self) -> Sequence[str]:
        return [
            self._external_sensor.handle.repository_handle.repository_name,
            self._external_sensor.name,
            self.tick_id,
        ]

    def update_state(self, status: TickStatus, **kwargs: object):
        skip_reason = cast(Optional[str], kwargs.get("skip_reason"))
        cursor = cast(Optional[str], kwargs.get("cursor"))
        origin_run_id = cast(Optional[str], kwargs.get("origin_run_id"))
        if "skip_reason" in kwargs:
            del kwargs["skip_reason"]

        if "cursor" in kwargs:
            del kwargs["cursor"]

        if "origin_run_id" in kwargs:
            del kwargs["origin_run_id"]
        if kwargs:
            check.inst_param(status, "status", TickStatus)

        if status:
            self._tick = self._tick.with_status(status=status, **kwargs)

        if skip_reason:
            self._tick = self._tick.with_reason(skip_reason=skip_reason)

        if cursor:
            self._tick = self._tick.with_cursor(cursor)

        if origin_run_id:
            self._tick = self._tick.with_origin_run(origin_run_id)

    def add_run_info(self, run_id: Optional[str] = None, run_key: Optional[str] = None) -> None:
        self._tick = self._tick.with_run_info(run_id, run_key)

    def add_log_key(self, log_key: Sequence[str]) -> None:
        self._tick = self._tick.with_log_key(log_key)

    def add_dynamic_partitions_request_result(
        self, dynamic_partitions_request_result: DynamicPartitionsRequestResult
    ) -> None:
        self._tick = self._tick.with_dynamic_partitions_request_result(
            dynamic_partitions_request_result
        )

    def set_should_update_cursor_on_failure(self, should_update_cursor_on_failure: bool) -> None:
        self._should_update_cursor_on_failure = should_update_cursor_on_failure

    def _write(self) -> None:
        self._instance.update_tick(self._tick)

        if self._tick.status not in FINISHED_TICK_STATES:
            return

        should_update_cursor_and_last_run_key = (
            self._tick.status != TickStatus.FAILURE
        ) or self._should_update_cursor_on_failure

        # fetch the most recent state.  we do this as opposed to during context initialization time
        # because we want to minimize the window of clobbering the sensor state upon updating the
        # sensor state data.
        state = self._instance.get_instigator_state(
            self._external_sensor.get_external_origin_id(), self._external_sensor.selector_id
        )
        last_run_key = state.instigator_data.last_run_key if state.instigator_data else None  # type: ignore  # (possible none)
        last_sensor_start_timestamp = (
            state.instigator_data.last_sensor_start_timestamp if state.instigator_data else None  # type: ignore  # (possible none)
        )
        if self._tick.run_keys and should_update_cursor_and_last_run_key:
            last_run_key = self._tick.run_keys[-1]

        cursor = state.instigator_data.cursor if state.instigator_data else None  # type: ignore  # (possible none)
        if should_update_cursor_and_last_run_key:
            cursor = self._tick.cursor

        marked_timestamp = max(
            self._tick.timestamp,
            state.instigator_data.last_tick_start_timestamp or 0,  # type: ignore  # (possible none)
        )
        self._instance.update_instigator_state(
            state.with_data(  # type: ignore  # (possible none)
                SensorInstigatorData(
                    last_tick_timestamp=self._tick.timestamp,
                    last_run_key=last_run_key,
                    min_interval=self._external_sensor.min_interval_seconds,
                    cursor=cursor,
                    last_tick_start_timestamp=marked_timestamp,
                    last_sensor_start_timestamp=last_sensor_start_timestamp,
                    sensor_type=self._external_sensor.sensor_type,
                )
            )
        )

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exception_type: Type[BaseException],
        exception_value: Exception,
        traceback: TracebackType,
    ) -> None:
        if exception_type and isinstance(exception_value, KeyboardInterrupt):
            return

        # Log the error if the failure wasn't an interrupt or the daemon generator stopping
        if exception_value and not isinstance(exception_value, GeneratorExit):
            error_info = DaemonErrorCapture.on_exception(exc_info=sys.exc_info())
            self.update_state(TickStatus.FAILURE, error=error_info)

        self._write()

        for day_offset, statuses in self._purge_settings.items():
            if day_offset <= 0:
                continue
            self._instance.purge_ticks(
                self._external_sensor.get_external_origin_id(),
                selector_id=self._external_sensor.selector_id,
                before=pendulum.now("UTC").subtract(days=day_offset).timestamp(),
                tick_statuses=list(statuses),
            )


def execute_sensor_iteration_loop(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    shutdown_event: threading.Event,
    until: Optional[float] = None,
    threadpool_executor: Optional[ThreadPoolExecutor] = None,
    submit_threadpool_executor: Optional[ThreadPoolExecutor] = None,
) -> "DaemonIterator":
    """Helper function that performs sensor evaluations on a tighter loop, while reusing grpc locations
    within a given daemon interval.  Rather than relying on the daemon machinery to run the
    iteration loop every 30 seconds, sensors are continuously evaluated, every 5 seconds. We rely on
    each sensor definition's min_interval to check that sensor evaluations are spaced appropriately.
    """
    from dagster._daemon.daemon import SpanMarker

    sensor_tick_futures: Dict[str, Future] = {}
    while True:
        start_time = pendulum.now("UTC").timestamp()
        if until and start_time >= until:
            # provide a way of organically ending the loop to support test environment
            break

        yield SpanMarker.START_SPAN

        try:
            yield from execute_sensor_iteration(
                workspace_process_context,
                logger,
                threadpool_executor=threadpool_executor,
                submit_threadpool_executor=submit_threadpool_executor,
                sensor_tick_futures=sensor_tick_futures,
            )
        except Exception:
            error_info = DaemonErrorCapture.on_exception(
                exc_info=sys.exc_info(),
                logger=logger,
                log_message="SensorDaemon caught an error",
            )
            yield error_info
        # Yield to check for heartbeats in case there were no yields within
        # execute_sensor_iteration
        yield SpanMarker.END_SPAN

        end_time = pendulum.now("UTC").timestamp()

        loop_duration = end_time - start_time
        sleep_time = max(0, MIN_INTERVAL_LOOP_TIME - loop_duration)
        shutdown_event.wait(sleep_time)

        yield None


def execute_sensor_iteration(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    threadpool_executor: Optional[ThreadPoolExecutor],
    submit_threadpool_executor: Optional[ThreadPoolExecutor],
    sensor_tick_futures: Optional[Dict[str, Future]] = None,
    debug_crash_flags: Optional[DebugCrashFlags] = None,
):
    instance = workspace_process_context.instance

    workspace_snapshot = {
        location_entry.origin.location_name: location_entry
        for location_entry in workspace_process_context.create_request_context()
        .get_workspace_snapshot()
        .values()
    }

    all_sensor_states = {
        sensor_state.selector_id: sensor_state
        for sensor_state in instance.all_instigator_state(instigator_type=InstigatorType.SENSOR)
        if (
            not sensor_state.instigator_data
            or sensor_state.instigator_data.sensor_type != SensorType.AUTO_MATERIALIZE  # type: ignore
        )
    }

    tick_retention_settings = instance.get_tick_retention_settings(InstigatorType.SENSOR)

    sensors: Dict[str, ExternalSensor] = {}
    for location_entry in workspace_snapshot.values():
        code_location = location_entry.code_location
        if code_location:
            for repo in code_location.get_repositories().values():
                for sensor in repo.get_external_sensors():
                    if sensor.sensor_type == SensorType.AUTO_MATERIALIZE:
                        continue

                    selector_id = sensor.selector_id
                    if sensor.get_current_instigator_state(
                        all_sensor_states.get(selector_id)
                    ).is_running:
                        sensors[selector_id] = sensor

    if not sensors:
        yield
        return

    for external_sensor in sensors.values():
        sensor_name = external_sensor.name
        sensor_debug_crash_flags = debug_crash_flags.get(sensor_name) if debug_crash_flags else None
        sensor_state = all_sensor_states.get(external_sensor.selector_id)
        if not sensor_state:
            assert external_sensor.default_status == DefaultSensorStatus.RUNNING
            sensor_state = InstigatorState(
                external_sensor.get_external_origin(),
                InstigatorType.SENSOR,
                InstigatorStatus.DECLARED_IN_CODE,
                SensorInstigatorData(
                    min_interval=external_sensor.min_interval_seconds,
                    last_sensor_start_timestamp=pendulum.now("UTC").timestamp(),
                    sensor_type=external_sensor.sensor_type,
                ),
            )
            instance.add_instigator_state(sensor_state)
        elif is_under_min_interval(sensor_state, external_sensor):
            continue

        if threadpool_executor:
            if sensor_tick_futures is None:
                check.failed("sensor_tick_futures dict must be passed with threadpool_executor")

            # only allow one tick per sensor to be in flight
            if (
                external_sensor.selector_id in sensor_tick_futures
                and not sensor_tick_futures[external_sensor.selector_id].done()
            ):
                continue

            future = threadpool_executor.submit(
                _process_tick,
                workspace_process_context,
                logger,
                external_sensor,
                sensor_state,
                sensor_debug_crash_flags,
                tick_retention_settings,
                submit_threadpool_executor,
            )
            sensor_tick_futures[external_sensor.selector_id] = future
            yield

        else:
            # evaluate the sensors in a loop, synchronously, yielding to allow the sensor daemon to
            # heartbeat
            yield from _process_tick_generator(
                workspace_process_context,
                logger,
                external_sensor,
                sensor_state,
                sensor_debug_crash_flags,
                tick_retention_settings,
                submit_threadpool_executor=None,
            )


def _process_tick(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    external_sensor: ExternalSensor,
    sensor_state: InstigatorState,
    sensor_debug_crash_flags: Optional[SingleInstigatorDebugCrashFlags],
    tick_retention_settings,
    submit_threadpool_executor: Optional[ThreadPoolExecutor],
):
    # evaluate the tick immediately, but from within a thread.  The main thread should be able to
    # heartbeat to keep the daemon alive
    return list(
        _process_tick_generator(
            workspace_process_context,
            logger,
            external_sensor,
            sensor_state,
            sensor_debug_crash_flags,
            tick_retention_settings,
            submit_threadpool_executor,
        )
    )


def _process_tick_generator(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    external_sensor: ExternalSensor,
    sensor_state: InstigatorState,
    sensor_debug_crash_flags: Optional[SingleInstigatorDebugCrashFlags],
    tick_retention_settings,
    submit_threadpool_executor: Optional[ThreadPoolExecutor],
):
    instance = workspace_process_context.instance
    error_info = None
    now = pendulum.now("UTC")
    sensor_state = check.not_none(
        instance.get_instigator_state(
            external_sensor.get_external_origin_id(), external_sensor.selector_id
        )
    )
    if is_under_min_interval(sensor_state, external_sensor):
        # check the since we might have been queued before processing
        return
    else:
        mark_sensor_state_for_tick(instance, external_sensor, sensor_state, now)

    try:
        tick = instance.create_tick(
            TickData(
                instigator_origin_id=sensor_state.instigator_origin_id,
                instigator_name=sensor_state.instigator_name,
                instigator_type=InstigatorType.SENSOR,
                status=TickStatus.STARTED,
                timestamp=now.timestamp(),
                selector_id=external_sensor.selector_id,
            )
        )

        check_for_debug_crash(sensor_debug_crash_flags, "TICK_CREATED")

        with SensorLaunchContext(
            external_sensor,
            tick,
            instance,
            logger,
            tick_retention_settings,
        ) as tick_context:
            check_for_debug_crash(sensor_debug_crash_flags, "TICK_HELD")
            tick_context.add_log_key(tick_context.log_key)

            yield from _evaluate_sensor(
                workspace_process_context,
                tick_context,
                external_sensor,
                sensor_state,
                submit_threadpool_executor,
                sensor_debug_crash_flags,
            )

    except Exception:
        error_info = DaemonErrorCapture.on_exception(
            exc_info=sys.exc_info(),
            logger=logger,
            log_message=f"Sensor daemon caught an error for sensor {external_sensor.name}",
        )

    yield error_info


def _sensor_instigator_data(state: InstigatorState) -> Optional[SensorInstigatorData]:
    instigator_data = state.instigator_data
    if instigator_data is None or isinstance(instigator_data, SensorInstigatorData):
        return instigator_data
    else:
        check.failed(f"Expected SensorInstigatorData, got {instigator_data}")


def mark_sensor_state_for_tick(
    instance: DagsterInstance,
    external_sensor: ExternalSensor,
    sensor_state: InstigatorState,
    now: "DateTime",
) -> None:
    instigator_data = _sensor_instigator_data(sensor_state)
    instance.update_instigator_state(
        sensor_state.with_data(
            SensorInstigatorData(
                last_tick_timestamp=(
                    instigator_data.last_tick_timestamp if instigator_data else None
                ),
                last_run_key=instigator_data.last_run_key if instigator_data else None,
                min_interval=external_sensor.min_interval_seconds,
                cursor=instigator_data.cursor if instigator_data else None,
                last_tick_start_timestamp=now.timestamp(),
                sensor_type=external_sensor.sensor_type,
                last_sensor_start_timestamp=instigator_data.last_sensor_start_timestamp
                if instigator_data
                else None,
            )
        )
    )


class SubmitRunRequestResult(NamedTuple):
    run_key: Optional[str]
    error_info: Optional[SerializableErrorInfo]
    run: Union[SkippedSensorRun, DagsterRun]


def _submit_run_request(
    run_request: RunRequest,
    workspace_process_context: IWorkspaceProcessContext,
    external_sensor: ExternalSensor,
    existing_runs_by_key,
    logger,
    sensor_debug_crash_flags,
) -> SubmitRunRequestResult:
    instance = workspace_process_context.instance
    sensor_origin = external_sensor.get_external_origin()

    target_data: ExternalTargetData = check.not_none(
        external_sensor.get_target_data(run_request.job_name)
    )

    # reload the code_location on each submission, request_context derived data can become out date
    # * non-threaded: if number of serial submissions is too many
    # * threaded: if thread sits pending in pool too long
    code_location = _get_code_location_for_sensor(workspace_process_context, external_sensor)
    job_subset_selector = JobSubsetSelector(
        location_name=code_location.name,
        repository_name=sensor_origin.repository_origin.repository_name,
        job_name=target_data.job_name,
        op_selection=target_data.op_selection,
        asset_selection=run_request.asset_selection,
        asset_check_selection=run_request.asset_check_keys,
    )
    external_job = code_location.get_external_job(job_subset_selector)
    run = _get_or_create_sensor_run(
        logger,
        instance,
        code_location,
        external_sensor,
        external_job,
        run_request,
        target_data,
        existing_runs_by_key,
    )

    if isinstance(run, SkippedSensorRun):
        return SubmitRunRequestResult(run_key=run_request.run_key, error_info=None, run=run)

    check_for_debug_crash(sensor_debug_crash_flags, "RUN_CREATED")

    error_info = None
    try:
        logger.info(f"Launching run for {external_sensor.name}")
        instance.submit_run(run.run_id, workspace_process_context.create_request_context())
        logger.info(f"Completed launch of run {run.run_id} for {external_sensor.name}")
    except Exception:
        error_info = DaemonErrorCapture.on_exception(
            exc_info=sys.exc_info(),
            logger=logger,
            log_message=f"Run {run.run_id} created successfully but failed to launch",
        )

    check_for_debug_crash(sensor_debug_crash_flags, "RUN_LAUNCHED")
    return SubmitRunRequestResult(run_key=run_request.run_key, error_info=error_info, run=run)


def _get_code_location_for_sensor(
    workspace_process_context: IWorkspaceProcessContext,
    external_sensor: ExternalSensor,
) -> CodeLocation:
    sensor_origin = external_sensor.get_external_origin()
    return workspace_process_context.create_request_context().get_code_location(
        sensor_origin.repository_origin.code_location_origin.location_name
    )


def _evaluate_sensor(
    workspace_process_context: IWorkspaceProcessContext,
    context: SensorLaunchContext,
    external_sensor: ExternalSensor,
    state: InstigatorState,
    submit_threadpool_executor: Optional[ThreadPoolExecutor],
    sensor_debug_crash_flags: Optional[SingleInstigatorDebugCrashFlags] = None,
):
    instance = workspace_process_context.instance
    context.logger.info(f"Checking for new runs for sensor: {external_sensor.name}")
    code_location = _get_code_location_for_sensor(workspace_process_context, external_sensor)
    repository_handle = external_sensor.handle.repository_handle
    instigator_data = _sensor_instigator_data(state)

    sensor_runtime_data = code_location.get_external_sensor_execution_data(
        instance,
        repository_handle,
        external_sensor.name,
        instigator_data.last_tick_timestamp if instigator_data else None,
        instigator_data.last_run_key if instigator_data else None,
        instigator_data.cursor if instigator_data else None,
        context.log_key,
        instigator_data.last_sensor_start_timestamp if instigator_data else None,
    )

    yield

    # Kept for backwards compatibility with sensor log keys that were previously created in the
    # sensor evaluation, rather than upfront.
    #
    # Note that to get sensor logs for failed sensor evaluations, we force users to update their
    # Dagster version.
    if sensor_runtime_data.log_key:
        context.add_log_key(sensor_runtime_data.log_key)

    for asset_event in sensor_runtime_data.asset_events:
        instance.report_runless_asset_event(asset_event)

    if sensor_runtime_data.dynamic_partitions_requests:
        _handle_dynamic_partitions_requests(
            sensor_runtime_data.dynamic_partitions_requests, instance, context
        )
    if not sensor_runtime_data.run_requests:
        if sensor_runtime_data.dagster_run_reactions:
            _handle_run_reactions(
                sensor_runtime_data.dagster_run_reactions,
                instance,
                context,
                sensor_runtime_data.cursor,
                external_sensor,
            )
        elif sensor_runtime_data.skip_message:
            context.logger.info(
                f"Sensor {external_sensor.name} skipped: {sensor_runtime_data.skip_message}"
            )
            context.update_state(
                TickStatus.SKIPPED,
                skip_reason=sensor_runtime_data.skip_message,
                cursor=sensor_runtime_data.cursor,
            )
        else:
            context.logger.info(f"No run requests returned for {external_sensor.name}, skipping")
            context.update_state(TickStatus.SKIPPED, cursor=sensor_runtime_data.cursor)

        yield
    else:
        yield from _handle_run_requests(
            run_requests=sensor_runtime_data.run_requests,
            cursor=sensor_runtime_data.cursor,
            context=context,
            instance=instance,
            external_sensor=external_sensor,
            workspace_process_context=workspace_process_context,
            submit_threadpool_executor=submit_threadpool_executor,
            sensor_debug_crash_flags=sensor_debug_crash_flags,
        )


def _handle_dynamic_partitions_requests(
    dynamic_partitions_requests: Sequence[
        Union[AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest]
    ],
    instance: DagsterInstance,
    context: SensorLaunchContext,
) -> None:
    for request in dynamic_partitions_requests:
        existent_partitions = []
        nonexistent_partitions = []
        for partition_key in request.partition_keys:
            if instance.has_dynamic_partition(request.partitions_def_name, partition_key):
                existent_partitions.append(partition_key)
            else:
                nonexistent_partitions.append(partition_key)

        if isinstance(request, AddDynamicPartitionsRequest):
            if nonexistent_partitions:
                instance.add_dynamic_partitions(
                    request.partitions_def_name,
                    nonexistent_partitions,
                )
                context.logger.info(
                    "Added partition keys to dynamic partitions definition"
                    f" '{request.partitions_def_name}': {nonexistent_partitions}"
                )

            if existent_partitions:
                context.logger.info(
                    "Skipping addition of partition keys for dynamic partitions definition"
                    f" '{request.partitions_def_name}' that already exist:"
                    f" {existent_partitions}"
                )

            context.add_dynamic_partitions_request_result(
                DynamicPartitionsRequestResult(
                    request.partitions_def_name,
                    added_partitions=nonexistent_partitions,
                    deleted_partitions=None,
                    skipped_partitions=existent_partitions,
                )
            )
        elif isinstance(request, DeleteDynamicPartitionsRequest):
            if existent_partitions:
                # TODO add a bulk delete method to the instance
                for partition in existent_partitions:
                    instance.delete_dynamic_partition(request.partitions_def_name, partition)

                context.logger.info(
                    "Deleted partition keys from dynamic partitions definition"
                    f" '{request.partitions_def_name}': {existent_partitions}"
                )

            if nonexistent_partitions:
                context.logger.info(
                    "Skipping deletion of partition keys for dynamic partitions definition"
                    f" '{request.partitions_def_name}' that do not exist:"
                    f" {nonexistent_partitions}"
                )

            context.add_dynamic_partitions_request_result(
                DynamicPartitionsRequestResult(
                    request.partitions_def_name,
                    added_partitions=None,
                    deleted_partitions=existent_partitions,
                    skipped_partitions=nonexistent_partitions,
                )
            )
        else:
            check.failed(f"Unexpected action {request.action} for dynamic partition request")


def _handle_run_reactions(
    dagster_run_reactions: Sequence[DagsterRunReaction],
    instance: DagsterInstance,
    context: SensorLaunchContext,
    cursor: Optional[str],
    external_sensor: ExternalSensor,
) -> None:
    for run_reaction in dagster_run_reactions:
        origin_run_id = check.not_none(run_reaction.dagster_run).run_id
        if run_reaction.error:
            context.logger.error(
                f"Got a reaction request for run {origin_run_id} but execution errorred:"
                f" {run_reaction.error}"
            )
            context.update_state(
                TickStatus.FAILURE,
                cursor=cursor,
                error=run_reaction.error,
            )
            # Since run status sensors have side effects that we don't want to repeat,
            # we still want to update the cursor, even though the tick failed
            context.set_should_update_cursor_on_failure(True)
        else:
            # Use status from the DagsterRunReaction object if it is from a new enough
            # version (0.14.4) to be set (the status on the DagsterRun object itself
            # may have since changed)
            status = (
                run_reaction.run_status.value
                if run_reaction.run_status
                else check.not_none(run_reaction.dagster_run).status.value
            )
            # log to the original dagster run
            message = (
                f'Sensor "{external_sensor.name}" acted on run status '
                f"{status} of run {origin_run_id}."
            )
            instance.report_engine_event(message=message, dagster_run=run_reaction.dagster_run)
            context.logger.info(f"Completed a reaction request for run {origin_run_id}: {message}")
            context.update_state(
                TickStatus.SUCCESS,
                cursor=cursor,
                origin_run_id=origin_run_id,
            )


def _handle_run_requests(
    run_requests: Sequence[RunRequest],
    instance: DagsterInstance,
    context: SensorLaunchContext,
    external_sensor: ExternalSensor,
    cursor: Optional[str],
    workspace_process_context: IWorkspaceProcessContext,
    submit_threadpool_executor: Optional[ThreadPoolExecutor],
    sensor_debug_crash_flags: Optional[SingleInstigatorDebugCrashFlags] = None,
):
    skipped_runs = []
    existing_runs_by_key = _fetch_existing_runs(instance, external_sensor, run_requests)

    resolved_run_requests = []

    for raw_run_request in run_requests:
        run_request = raw_run_request.with_replaced_attrs(
            tags=merge_dicts(
                raw_run_request.tags,
                DagsterRun.tags_for_tick_id(context.tick_id),
            )
        )

        if run_request.stale_assets_only:
            stale_assets = resolve_stale_or_missing_assets(
                workspace_process_context,  # type: ignore
                run_request,
                external_sensor,
            )
            # asset selection is empty set after filtering for stale
            if len(stale_assets) == 0:
                continue
            else:
                run_request = run_request.with_replaced_attrs(
                    asset_selection=stale_assets, stale_assets_only=False
                )

        resolved_run_requests.append(run_request)

    def submit_run_request(run_request) -> SubmitRunRequestResult:
        return _submit_run_request(
            run_request,
            workspace_process_context,
            external_sensor,
            existing_runs_by_key,
            context.logger,
            sensor_debug_crash_flags,
        )

    if submit_threadpool_executor:
        gen_run_request_results = submit_threadpool_executor.map(
            submit_run_request, resolved_run_requests
        )
    else:
        gen_run_request_results = map(submit_run_request, resolved_run_requests)

    for run_request_result in gen_run_request_results:
        yield run_request_result.error_info

        run = run_request_result.run

        if isinstance(run, SkippedSensorRun):
            skipped_runs.append(run)
            context.add_run_info(run_id=None, run_key=run_request_result.run_key)
        else:
            context.add_run_info(run_id=run.run_id, run_key=run_request_result.run_key)

    if skipped_runs:
        run_keys = [skipped.run_key for skipped in skipped_runs]
        skipped_count = len(skipped_runs)
        context.logger.info(
            f"Skipping {skipped_count} {'run' if skipped_count == 1 else 'runs'} for sensor "
            f"{external_sensor.name} already completed with run keys: {seven.json.dumps(run_keys)}"
        )

    if context.run_count:
        context.update_state(TickStatus.SUCCESS, cursor=cursor)
    else:
        context.update_state(TickStatus.SKIPPED, cursor=cursor)

    yield


def is_under_min_interval(state: InstigatorState, external_sensor: ExternalSensor) -> bool:
    instigator_data = _sensor_instigator_data(state)
    if not instigator_data:
        return False

    if not instigator_data.last_tick_start_timestamp and not instigator_data.last_tick_timestamp:
        return False

    if not external_sensor.min_interval_seconds:
        return False

    elapsed = pendulum.now("UTC").timestamp() - max(
        instigator_data.last_tick_timestamp or 0,
        instigator_data.last_tick_start_timestamp or 0,
    )
    return elapsed < external_sensor.min_interval_seconds


def _fetch_existing_runs(
    instance: DagsterInstance,
    external_sensor: ExternalSensor,
    run_requests: Sequence[RunRequest],
):
    run_keys = [run_request.run_key for run_request in run_requests if run_request.run_key]

    if not run_keys:
        return {}

    # fetch runs from the DB with only the run key tag
    # note: while possible to filter more at DB level with tags - it is avoided here due to observed
    # perf problems
    runs_with_run_keys = []
    for run_key in run_keys:
        # do serial fetching, which has better perf than a single query with an IN clause, due to
        # how the query planner does the runs/run_tags join
        runs_with_run_keys.extend(
            instance.get_runs(filters=RunsFilter(tags={RUN_KEY_TAG: run_key}))
        )

    # filter down to runs with run_key that match the sensor name and its namespace (repository)
    valid_runs: List[DagsterRun] = []
    for run in runs_with_run_keys:
        # if the run doesn't have a set origin, just match on sensor name
        if (
            run.external_job_origin is None
            and run.tags.get(SENSOR_NAME_TAG) == external_sensor.name
        ):
            valid_runs.append(run)
        # otherwise prevent the same named sensor across repos from effecting each other
        elif (
            run.external_job_origin is not None
            and run.external_job_origin.repository_origin.get_selector_id()
            == external_sensor.get_external_origin().repository_origin.get_selector_id()
            and run.tags.get(SENSOR_NAME_TAG) == external_sensor.name
        ):
            valid_runs.append(run)

    existing_runs = {}
    for run in valid_runs:
        tags = run.tags or {}
        run_key = tags.get(RUN_KEY_TAG)
        existing_runs[run_key] = run

    return existing_runs


def _get_or_create_sensor_run(
    logger: logging.Logger,
    instance: DagsterInstance,
    code_location: CodeLocation,
    external_sensor: ExternalSensor,
    external_job: ExternalJob,
    run_request: RunRequest,
    target_data: ExternalTargetData,
    existing_runs_by_key: Mapping[str, DagsterRun],
) -> Union[DagsterRun, SkippedSensorRun]:
    if not run_request.run_key:
        return _create_sensor_run(
            instance, code_location, external_sensor, external_job, run_request, target_data
        )

    run = existing_runs_by_key.get(run_request.run_key)

    if run:
        if run.status != DagsterRunStatus.NOT_STARTED:
            # A run already exists and was launched for this run key, but the daemon must have
            # crashed before the tick could be updated
            return SkippedSensorRun(run_key=run_request.run_key, existing_run=run)
        else:
            logger.info(
                f"Run {run.run_id} already created with the run key "
                f"`{run_request.run_key}` for {external_sensor.name}"
            )
            return run

    logger.info(f"Creating new run for {external_sensor.name}")

    return _create_sensor_run(
        instance, code_location, external_sensor, external_job, run_request, target_data
    )


def _create_sensor_run(
    instance: DagsterInstance,
    code_location: CodeLocation,
    external_sensor: ExternalSensor,
    external_job: ExternalJob,
    run_request: RunRequest,
    target_data: ExternalTargetData,
) -> DagsterRun:
    from dagster._daemon.daemon import get_telemetry_daemon_session_id

    external_execution_plan = code_location.get_external_execution_plan(
        external_job,
        run_request.run_config,
        step_keys_to_execute=None,
        known_state=None,
        instance=instance,
    )
    execution_plan_snapshot = external_execution_plan.execution_plan_snapshot

    job_tags = normalize_tags(
        external_job.tags or {}, allow_reserved_tags=False, warn_on_deprecated_tags=False
    ).tags
    tags = merge_dicts(
        merge_dicts(job_tags, run_request.tags),
        # this gets applied in the sensor definition too, but we apply it here for backcompat
        # with sensors before the tag was added to the sensor definition
        DagsterRun.tags_for_sensor(external_sensor),
    )
    if run_request.run_key:
        tags[RUN_KEY_TAG] = run_request.run_key

    log_action(
        instance,
        SENSOR_RUN_CREATED,
        metadata={
            "DAEMON_SESSION_ID": get_telemetry_daemon_session_id(),
            "SENSOR_NAME_HASH": hash_name(external_sensor.name),
            "pipeline_name_hash": hash_name(external_job.name),
            "repo_hash": hash_name(code_location.name),
        },
    )

    return instance.create_run(
        job_name=target_data.job_name,
        run_id=None,
        run_config=run_request.run_config,
        resolved_op_selection=external_job.resolved_op_selection,
        step_keys_to_execute=None,
        status=DagsterRunStatus.NOT_STARTED,
        op_selection=target_data.op_selection,
        root_run_id=None,
        parent_run_id=None,
        tags=tags,
        job_snapshot=external_job.job_snapshot,
        execution_plan_snapshot=execution_plan_snapshot,
        parent_job_snapshot=external_job.parent_job_snapshot,
        external_job_origin=external_job.get_external_origin(),
        job_code_origin=external_job.get_python_origin(),
        asset_selection=(
            frozenset(run_request.asset_selection) if run_request.asset_selection else None
        ),
        asset_check_selection=(
            frozenset(run_request.asset_check_keys) if run_request.asset_check_keys else None
        ),
        asset_job_partitions_def=code_location.get_asset_job_partitions_def(external_job),
    )



from typing import (
    Any,
    Mapping,
    Optional,
    Sequence,
)

from dagster import _check as check
from dagster._annotations import deprecated
from dagster._core.definitions.asset_check_spec import AssetCheckSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.errors import DagsterAssetCheckFailedError
from dagster._core.types.dagster_type import Nothing


class AssetChecksDefinition(AssetsDefinition):
    """Defines a set of checks that are produced by the same op or op graph.

    AssetChecksDefinition should not be instantiated directly, but rather produced using the `@asset_check` decorator or `AssetChecksDefinition.create` method.
    """

    @staticmethod
    def create(
        *,
        keys_by_input_name: Mapping[str, AssetKey],
        node_def: OpDefinition,
        check_specs_by_output_name: Mapping[str, AssetCheckSpec],
        can_subset: bool,
        resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
    ):
        """Create an AssetChecksDefinition."""
        return AssetChecksDefinition(
            keys_by_input_name=keys_by_input_name,
            keys_by_output_name={},
            node_def=node_def,
            partitions_def=None,
            partition_mappings=None,
            asset_deps=None,
            selected_asset_keys=None,
            can_subset=can_subset,
            resource_defs=resource_defs,
            group_names_by_key=None,
            metadata_by_key=None,
            tags_by_key=None,
            freshness_policies_by_key=None,
            auto_materialize_policies_by_key=None,
            backfill_policy=None,
            descriptions_by_key=None,
            check_specs_by_output_name=check_specs_by_output_name,
            selected_asset_check_keys=None,
            is_subset=False,
            owners_by_key=None,
        )


# This is still needed in a few places where we need to handle normal AssetsDefinition and
# AssetChecksDefinition differently, but eventually those areas should be refactored and this should
# be removed.
def has_only_asset_checks(assets_def: AssetsDefinition) -> bool:
    return len(assets_def.keys) == 0 and len(assets_def.check_keys) > 0


@deprecated(breaking_version="1.8", additional_warn_text="Use @asset_check(blocking=True) instead")
def build_asset_with_blocking_check(
    asset_def: "AssetsDefinition",
    checks: Sequence["AssetsDefinition"],
) -> "AssetsDefinition":
    from dagster import AssetIn, In, OpExecutionContext, Output, op
    from dagster._core.definitions.decorators.asset_decorator import graph_asset_no_defaults
    from dagster._core.storage.asset_check_execution_record import AssetCheckExecutionRecordStatus

    check_specs = []
    for c in checks:
        check_specs.extend(c.check_specs)

    check_output_names = [c.get_python_identifier() for c in check_specs]

    check.invariant(len(asset_def.op.output_defs) == 1)
    asset_out_type = asset_def.op.output_defs[0].dagster_type

    @op(
        name=f"{asset_def.op.name}_asset_and_checks",
        ins={"asset_return_value": In(asset_out_type), "check_evaluations": In(Nothing)},
    )
    def fan_in_checks_and_asset_return_value(context: OpExecutionContext, asset_return_value: Any):
        # we pass the asset_return_value through and store it again so that downstream assets can load it.
        # This is a little silly- we only do this because this op has the asset key in its StepOutputProperties
        # so the output is written to the right path. We could probably get the asset_def.op to write to the
        # asset path (and make sure we don't override it here) to avoid the double write.
        yield Output(asset_return_value)

        for check_spec in check_specs:
            executions = context.instance.event_log_storage.get_asset_check_execution_history(
                check_key=check_spec.key, limit=1
            )
            check.invariant(
                len(executions) == 1, "Expected asset check {check_spec.name} to execute"
            )
            execution = executions[0]
            check.invariant(
                execution.run_id == context.run_id,
                "Expected asset check {check_spec.name} to execute in the current run",
            )
            if execution.status != AssetCheckExecutionRecordStatus.SUCCEEDED:
                raise DagsterAssetCheckFailedError()

    # kwargs are the inputs to the asset_def.op that we are wrapping
    def blocking_asset(**kwargs):
        asset_return_value = asset_def.op.with_replaced_properties(
            name=f"{asset_def.op.name}_graph_asset_op"
        )(**kwargs)
        check_evaluations = [check.node_def(asset_return_value) for check in checks]

        return {
            "result": fan_in_checks_and_asset_return_value(asset_return_value, check_evaluations),
            **{
                check_output_name: check_result
                for check_output_name, check_result in zip(check_output_names, check_evaluations)
            },
        }

    return graph_asset_no_defaults(
        compose_fn=blocking_asset,
        name=None,
        key_prefix=None,
        key=asset_def.key,
        group_name=asset_def.group_names_by_key.get(asset_def.key),
        partitions_def=asset_def.partitions_def,
        check_specs=check_specs,
        description=asset_def.descriptions_by_key.get(asset_def.key),
        ins={name: AssetIn(key) for name, key in asset_def.keys_by_input_name.items()},
        resource_defs=asset_def.resource_defs,
        metadata=asset_def.metadata_by_key.get(asset_def.key),
        tags=None,
        owners=None,
        freshness_policy=asset_def.freshness_policies_by_key.get(asset_def.key),
        auto_materialize_policy=asset_def.auto_materialize_policies_by_key.get(asset_def.key),
        backfill_policy=asset_def.backfill_policy,
        config=None,  # gets config from asset_def.op
    )



from .parent_newer_condition import ParentNewerCondition as ParentNewerCondition
from .scheduled_since_condition import ScheduledSinceCondition as ScheduledSinceCondition
from .slice_conditions import (
    FailedSchedulingCondition as FailedSchedulingCondition,
    InLatestTimeWindowCondition as InLatestTimeWindowCondition,
    InProgressSchedulingCondition as InProgressSchedulingCondition,
    MissingSchedulingCondition as MissingSchedulingCondition,
    RequestedThisTickCondition as RequestedThisTickCondition,
)
from .updated_since_cron_condition import UpdatedSinceCronCondition as UpdatedSinceCronCondition



import datetime

from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.schedules import reverse_cron_string_iterator

from ..scheduling_condition import SchedulingCondition, SchedulingResult
from ..scheduling_context import SchedulingContext


@whitelist_for_serdes
class UpdatedSinceCronCondition(SchedulingCondition):
    cron_schedule: str
    cron_timezone: str

    @property
    def description(self) -> str:
        return f"Updated since latest tick of {self.cron_schedule} ({self.cron_timezone})"

    def _get_previous_cron_tick(self, context: SchedulingContext) -> datetime.datetime:
        previous_ticks = reverse_cron_string_iterator(
            end_timestamp=context.effective_dt.timestamp(),
            cron_string=self.cron_schedule,
            execution_timezone=self.cron_timezone,
        )
        return next(previous_ticks)

    def evaluate(self, context: SchedulingContext) -> SchedulingResult:
        previous_cron_tick = self._get_previous_cron_tick(context)

        if (
            # never evaluated or evaluation info is no longer valid
            context.previous_evaluation_info is None
            or context.previous_evaluation_node is None
            or context.previous_evaluation_node.true_slice is None
            # not evaluated since latest schedule tick
            or context.previous_evaluation_info.temporal_context.effective_dt < previous_cron_tick
            # has new set of candidates
            or context.previous_evaluation_node.candidate_slice != context.candidate_slice
        ):
            # do a full recomputation
            updated_subset = (
                context.legacy_context.instance_queryer.get_asset_subset_updated_after_time(
                    asset_key=context.asset_key,
                    after_time=previous_cron_tick,
                )
            )
            # TODO: implement this on the AssetGraphView
            true_slice = context.asset_graph_view.get_asset_slice_from_valid_subset(updated_subset)
        else:
            # do an incremental updated, adding in any asset partitions that have been materialized
            # since the previous evaluation
            true_slice = context.previous_evaluation_node.true_slice.compute_union(
                context.asset_graph_view.compute_updated_since_cursor_slice(
                    asset_key=context.asset_key, cursor=context.previous_evaluation_max_storage_id
                )
            )

        return SchedulingResult.create(
            context, context.candidate_slice.compute_intersection(true_slice)
        )



from dagster._core.definitions.asset_key import AssetKey
from dagster._serdes.serdes import whitelist_for_serdes

from ..scheduling_condition import SchedulingCondition, SchedulingResult
from ..scheduling_context import SchedulingContext


class DepConditionWrapperCondition(SchedulingCondition):
    """Wrapper object which evaluates a condition against a dependency and returns a subset
    representing the subset of downstream asset which has at least one parent which evaluated to
    True.
    """

    dep_key: AssetKey
    operand: SchedulingCondition

    @property
    def description(self) -> str:
        return f"{self.dep_key.to_user_string()}"

    def evaluate(self, context: SchedulingContext) -> SchedulingResult:
        # only evaluate parents of the current candidates
        dep_candidate_slice = context.candidate_slice.compute_parent_slice(self.dep_key)
        dep_context = context.for_child_condition(
            child_condition=self.operand, child_index=0, candidate_slice=dep_candidate_slice
        )

        # evaluate condition against the dependency
        dep_result = self.operand.evaluate(dep_context)

        # find all children of the true dep slice
        true_slice = dep_result.true_slice.compute_child_slice(context.asset_key)
        return SchedulingResult.create_from_children(
            context=context, true_slice=true_slice, child_results=[dep_result]
        )


class DepCondition(SchedulingCondition):
    operand: SchedulingCondition


@whitelist_for_serdes
class AnyDepsCondition(DepCondition):
    @property
    def description(self) -> str:
        return "Any deps"

    def evaluate(self, context: SchedulingContext) -> SchedulingResult:
        dep_results = []
        true_slice = context.asset_graph_view.create_empty_slice(context.asset_key)

        dep_keys = context.asset_graph_view.asset_graph.get(context.asset_key).parent_keys
        for i, dep_key in enumerate(sorted(dep_keys)):
            dep_condition = DepConditionWrapperCondition(dep_key=dep_key, operand=self.operand)
            dep_result = dep_condition.evaluate(
                context.for_child_condition(
                    child_condition=dep_condition,
                    child_index=i,
                    candidate_slice=context.candidate_slice,
                )
            )
            dep_results.append(dep_result)
            true_slice = true_slice.compute_union(dep_result.true_slice)

        true_slice = context.candidate_slice.compute_intersection(true_slice)
        return SchedulingResult.create_from_children(
            context, true_slice=true_slice, child_results=dep_results
        )


@whitelist_for_serdes
class AllDepsCondition(DepCondition):
    @property
    def description(self) -> str:
        return "All deps"

    def evaluate(self, context: SchedulingContext) -> SchedulingResult:
        dep_results = []
        true_slice = context.candidate_slice

        dep_keys = context.asset_graph_view.asset_graph.get(context.asset_key).parent_keys
        for i, dep_key in enumerate(sorted(dep_keys)):
            dep_condition = DepConditionWrapperCondition(dep_key=dep_key, operand=self.operand)
            dep_result = dep_condition.evaluate(
                context.for_child_condition(
                    child_condition=dep_condition,
                    child_index=i,
                    candidate_slice=context.candidate_slice,
                )
            )
            dep_results.append(dep_result)
            true_slice = true_slice.compute_intersection(dep_result.true_slice)

        return SchedulingResult.create_from_children(
            context, true_slice=true_slice, child_results=dep_results
        )



# @generated

# This file was generated by running `python -m dagster._grpc.compile`
# Do not edit this file directly, and do not attempt to recompile it using
# grpc_tools.protoc directly, as several changes must be made to the raw output

# noqa: SLF001

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\tapi.proto\x12\x03\x61pi"\x07\n\x05\x45mpty"\x1b\n\x0bPingRequest\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t"H\n\tPingReply\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t\x12-\n%serialized_server_utilization_metrics\x18\x02 \x01(\t"=\n\x14StreamingPingRequest\x12\x17\n\x0fsequence_length\x18\x01 \x01(\x05\x12\x0c\n\x04\x65\x63ho\x18\x02 \x01(\t";\n\x12StreamingPingEvent\x12\x17\n\x0fsequence_number\x18\x01 \x01(\x05\x12\x0c\n\x04\x65\x63ho\x18\x02 \x01(\t"%\n\x10GetServerIdReply\x12\x11\n\tserver_id\x18\x01 \x01(\t"O\n\x1c\x45xecutionPlanSnapshotRequest\x12/\n\'serialized_execution_plan_snapshot_args\x18\x01 \x01(\t"H\n\x1a\x45xecutionPlanSnapshotReply\x12*\n"serialized_execution_plan_snapshot\x18\x01 \x01(\t"H\n\x1d\x45xternalPartitionNamesRequest\x12\'\n\x1fserialized_partition_names_args\x18\x01 \x01(\t"p\n\x1b\x45xternalPartitionNamesReply\x12Q\nIserialized_external_partition_names_or_external_partition_execution_error\x18\x01 \x01(\t"4\n\x1b\x45xternalNotebookDataRequest\x12\x15\n\rnotebook_path\x18\x01 \x01(\t",\n\x19\x45xternalNotebookDataReply\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c"C\n\x1e\x45xternalPartitionConfigRequest\x12!\n\x19serialized_partition_args\x18\x01 \x01(\t"r\n\x1c\x45xternalPartitionConfigReply\x12R\nJserialized_external_partition_config_or_external_partition_execution_error\x18\x01 \x01(\t"A\n\x1c\x45xternalPartitionTagsRequest\x12!\n\x19serialized_partition_args\x18\x01 \x01(\t"n\n\x1a\x45xternalPartitionTagsReply\x12P\nHserialized_external_partition_tags_or_external_partition_execution_error\x18\x01 \x01(\t"c\n*ExternalPartitionSetExecutionParamsRequest\x12\x35\n-serialized_partition_set_execution_param_args\x18\x01 \x01(\t"\x19\n\x17ListRepositoriesRequest"O\n\x15ListRepositoriesReply\x12\x36\n.serialized_list_repositories_response_or_error\x18\x01 \x01(\t"Y\n%ExternalPipelineSubsetSnapshotRequest\x12\x30\n(serialized_pipeline_subset_snapshot_args\x18\x01 \x01(\t"Y\n#ExternalPipelineSubsetSnapshotReply\x12\x32\n*serialized_external_pipeline_subset_result\x18\x01 \x01(\t"a\n\x19\x45xternalRepositoryRequest\x12+\n#serialized_repository_python_origin\x18\x01 \x01(\t\x12\x17\n\x0f\x64\x65\x66\x65r_snapshots\x18\x02 \x01(\x08"F\n\x17\x45xternalRepositoryReply\x12+\n#serialized_external_repository_data\x18\x01 \x01(\t"i\n StreamingExternalRepositoryEvent\x12\x17\n\x0fsequence_number\x18\x01 \x01(\x05\x12,\n$serialized_external_repository_chunk\x18\x02 \x01(\t"W\n ExternalScheduleExecutionRequest\x12\x33\n+serialized_external_schedule_execution_args\x18\x01 \x01(\t"S\n\x1e\x45xternalSensorExecutionRequest\x12\x31\n)serialized_external_sensor_execution_args\x18\x01 \x01(\t"H\n\x13StreamingChunkEvent\x12\x17\n\x0fsequence_number\x18\x01 \x01(\x05\x12\x18\n\x10serialized_chunk\x18\x02 \x01(\t"@\n\x13ShutdownServerReply\x12)\n!serialized_shutdown_server_result\x18\x01 \x01(\t"E\n\x16\x43\x61ncelExecutionRequest\x12+\n#serialized_cancel_execution_request\x18\x01 \x01(\t"B\n\x14\x43\x61ncelExecutionReply\x12*\n"serialized_cancel_execution_result\x18\x01 \x01(\t"L\n\x19\x43\x61nCancelExecutionRequest\x12/\n\'serialized_can_cancel_execution_request\x18\x01 \x01(\t"I\n\x17\x43\x61nCancelExecutionReply\x12.\n&serialized_can_cancel_execution_result\x18\x01 \x01(\t"6\n\x0fStartRunRequest\x12#\n\x1bserialized_execute_run_args\x18\x01 \x01(\t"4\n\rStartRunReply\x12#\n\x1bserialized_start_run_result\x18\x01 \x01(\t"8\n\x14GetCurrentImageReply\x12 \n\x18serialized_current_image\x18\x01 \x01(\t"6\n\x13GetCurrentRunsReply\x12\x1f\n\x17serialized_current_runs\x18\x01 \x01(\t"L\n\x12\x45xternalJobRequest\x12$\n\x1cserialized_repository_origin\x18\x01 \x01(\t\x12\x10\n\x08job_name\x18\x02 \x01(\t"I\n\x10\x45xternalJobReply\x12\x1b\n\x13serialized_job_data\x18\x01 \x01(\t\x12\x18\n\x10serialized_error\x18\x02 \x01(\t"D\n\x1e\x45xternalScheduleExecutionReply\x12"\n\x1aserialized_schedule_result\x18\x01 \x01(\t"@\n\x1c\x45xternalSensorExecutionReply\x12 \n\x18serialized_sensor_result\x18\x01 \x01(\t"\x13\n\x11ReloadCodeRequest"+\n\x0fReloadCodeReply\x12\x18\n\x10serialized_error\x18\x02 \x01(\t2\xe9\x10\n\nDagsterApi\x12*\n\x04Ping\x12\x10.api.PingRequest\x1a\x0e.api.PingReply"\x00\x12/\n\tHeartbeat\x12\x10.api.PingRequest\x1a\x0e.api.PingReply"\x00\x12G\n\rStreamingPing\x12\x19.api.StreamingPingRequest\x1a\x17.api.StreamingPingEvent"\x00\x30\x01\x12\x32\n\x0bGetServerId\x12\n.api.Empty\x1a\x15.api.GetServerIdReply"\x00\x12]\n\x15\x45xecutionPlanSnapshot\x12!.api.ExecutionPlanSnapshotRequest\x1a\x1f.api.ExecutionPlanSnapshotReply"\x00\x12N\n\x10ListRepositories\x12\x1c.api.ListRepositoriesRequest\x1a\x1a.api.ListRepositoriesReply"\x00\x12`\n\x16\x45xternalPartitionNames\x12".api.ExternalPartitionNamesRequest\x1a .api.ExternalPartitionNamesReply"\x00\x12Z\n\x14\x45xternalNotebookData\x12 .api.ExternalNotebookDataRequest\x1a\x1e.api.ExternalNotebookDataReply"\x00\x12\x63\n\x17\x45xternalPartitionConfig\x12#.api.ExternalPartitionConfigRequest\x1a!.api.ExternalPartitionConfigReply"\x00\x12]\n\x15\x45xternalPartitionTags\x12!.api.ExternalPartitionTagsRequest\x1a\x1f.api.ExternalPartitionTagsReply"\x00\x12t\n#ExternalPartitionSetExecutionParams\x12/.api.ExternalPartitionSetExecutionParamsRequest\x1a\x18.api.StreamingChunkEvent"\x00\x30\x01\x12x\n\x1e\x45xternalPipelineSubsetSnapshot\x12*.api.ExternalPipelineSubsetSnapshotRequest\x1a(.api.ExternalPipelineSubsetSnapshotReply"\x00\x12T\n\x12\x45xternalRepository\x12\x1e.api.ExternalRepositoryRequest\x1a\x1c.api.ExternalRepositoryReply"\x00\x12?\n\x0b\x45xternalJob\x12\x17.api.ExternalJobRequest\x1a\x15.api.ExternalJobReply"\x00\x12h\n\x1bStreamingExternalRepository\x12\x1e.api.ExternalRepositoryRequest\x1a%.api.StreamingExternalRepositoryEvent"\x00\x30\x01\x12`\n\x19\x45xternalScheduleExecution\x12%.api.ExternalScheduleExecutionRequest\x1a\x18.api.StreamingChunkEvent"\x00\x30\x01\x12m\n\x1dSyncExternalScheduleExecution\x12%.api.ExternalScheduleExecutionRequest\x1a#.api.ExternalScheduleExecutionReply"\x00\x12\\\n\x17\x45xternalSensorExecution\x12#.api.ExternalSensorExecutionRequest\x1a\x18.api.StreamingChunkEvent"\x00\x30\x01\x12g\n\x1bSyncExternalSensorExecution\x12#.api.ExternalSensorExecutionRequest\x1a!.api.ExternalSensorExecutionReply"\x00\x12\x38\n\x0eShutdownServer\x12\n.api.Empty\x1a\x18.api.ShutdownServerReply"\x00\x12K\n\x0f\x43\x61ncelExecution\x12\x1b.api.CancelExecutionRequest\x1a\x19.api.CancelExecutionReply"\x00\x12T\n\x12\x43\x61nCancelExecution\x12\x1e.api.CanCancelExecutionRequest\x1a\x1c.api.CanCancelExecutionReply"\x00\x12\x36\n\x08StartRun\x12\x14.api.StartRunRequest\x1a\x12.api.StartRunReply"\x00\x12:\n\x0fGetCurrentImage\x12\n.api.Empty\x1a\x19.api.GetCurrentImageReply"\x00\x12\x38\n\x0eGetCurrentRuns\x12\n.api.Empty\x1a\x18.api.GetCurrentRunsReply"\x00\x12<\n\nReloadCode\x12\x16.api.ReloadCodeRequest\x1a\x14.api.ReloadCodeReply"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "api_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_EMPTY"]._serialized_start = 18
    _globals["_EMPTY"]._serialized_end = 25
    _globals["_PINGREQUEST"]._serialized_start = 27
    _globals["_PINGREQUEST"]._serialized_end = 54
    _globals["_PINGREPLY"]._serialized_start = 56
    _globals["_PINGREPLY"]._serialized_end = 128
    _globals["_STREAMINGPINGREQUEST"]._serialized_start = 130
    _globals["_STREAMINGPINGREQUEST"]._serialized_end = 191
    _globals["_STREAMINGPINGEVENT"]._serialized_start = 193
    _globals["_STREAMINGPINGEVENT"]._serialized_end = 252
    _globals["_GETSERVERIDREPLY"]._serialized_start = 254
    _globals["_GETSERVERIDREPLY"]._serialized_end = 291
    _globals["_EXECUTIONPLANSNAPSHOTREQUEST"]._serialized_start = 293
    _globals["_EXECUTIONPLANSNAPSHOTREQUEST"]._serialized_end = 372
    _globals["_EXECUTIONPLANSNAPSHOTREPLY"]._serialized_start = 374
    _globals["_EXECUTIONPLANSNAPSHOTREPLY"]._serialized_end = 446
    _globals["_EXTERNALPARTITIONNAMESREQUEST"]._serialized_start = 448
    _globals["_EXTERNALPARTITIONNAMESREQUEST"]._serialized_end = 520
    _globals["_EXTERNALPARTITIONNAMESREPLY"]._serialized_start = 522
    _globals["_EXTERNALPARTITIONNAMESREPLY"]._serialized_end = 634
    _globals["_EXTERNALNOTEBOOKDATAREQUEST"]._serialized_start = 636
    _globals["_EXTERNALNOTEBOOKDATAREQUEST"]._serialized_end = 688
    _globals["_EXTERNALNOTEBOOKDATAREPLY"]._serialized_start = 690
    _globals["_EXTERNALNOTEBOOKDATAREPLY"]._serialized_end = 734
    _globals["_EXTERNALPARTITIONCONFIGREQUEST"]._serialized_start = 736
    _globals["_EXTERNALPARTITIONCONFIGREQUEST"]._serialized_end = 803
    _globals["_EXTERNALPARTITIONCONFIGREPLY"]._serialized_start = 805
    _globals["_EXTERNALPARTITIONCONFIGREPLY"]._serialized_end = 919
    _globals["_EXTERNALPARTITIONTAGSREQUEST"]._serialized_start = 921
    _globals["_EXTERNALPARTITIONTAGSREQUEST"]._serialized_end = 986
    _globals["_EXTERNALPARTITIONTAGSREPLY"]._serialized_start = 988
    _globals["_EXTERNALPARTITIONTAGSREPLY"]._serialized_end = 1098
    _globals["_EXTERNALPARTITIONSETEXECUTIONPARAMSREQUEST"]._serialized_start = 1100
    _globals["_EXTERNALPARTITIONSETEXECUTIONPARAMSREQUEST"]._serialized_end = 1199
    _globals["_LISTREPOSITORIESREQUEST"]._serialized_start = 1201
    _globals["_LISTREPOSITORIESREQUEST"]._serialized_end = 1226
    _globals["_LISTREPOSITORIESREPLY"]._serialized_start = 1228
    _globals["_LISTREPOSITORIESREPLY"]._serialized_end = 1307
    _globals["_EXTERNALPIPELINESUBSETSNAPSHOTREQUEST"]._serialized_start = 1309
    _globals["_EXTERNALPIPELINESUBSETSNAPSHOTREQUEST"]._serialized_end = 1398
    _globals["_EXTERNALPIPELINESUBSETSNAPSHOTREPLY"]._serialized_start = 1400
    _globals["_EXTERNALPIPELINESUBSETSNAPSHOTREPLY"]._serialized_end = 1489
    _globals["_EXTERNALREPOSITORYREQUEST"]._serialized_start = 1491
    _globals["_EXTERNALREPOSITORYREQUEST"]._serialized_end = 1588
    _globals["_EXTERNALREPOSITORYREPLY"]._serialized_start = 1590
    _globals["_EXTERNALREPOSITORYREPLY"]._serialized_end = 1660
    _globals["_STREAMINGEXTERNALREPOSITORYEVENT"]._serialized_start = 1662
    _globals["_STREAMINGEXTERNALREPOSITORYEVENT"]._serialized_end = 1767
    _globals["_EXTERNALSCHEDULEEXECUTIONREQUEST"]._serialized_start = 1769
    _globals["_EXTERNALSCHEDULEEXECUTIONREQUEST"]._serialized_end = 1856
    _globals["_EXTERNALSENSOREXECUTIONREQUEST"]._serialized_start = 1858
    _globals["_EXTERNALSENSOREXECUTIONREQUEST"]._serialized_end = 1941
    _globals["_STREAMINGCHUNKEVENT"]._serialized_start = 1943
    _globals["_STREAMINGCHUNKEVENT"]._serialized_end = 2015
    _globals["_SHUTDOWNSERVERREPLY"]._serialized_start = 2017
    _globals["_SHUTDOWNSERVERREPLY"]._serialized_end = 2081
    _globals["_CANCELEXECUTIONREQUEST"]._serialized_start = 2083
    _globals["_CANCELEXECUTIONREQUEST"]._serialized_end = 2152
    _globals["_CANCELEXECUTIONREPLY"]._serialized_start = 2154
    _globals["_CANCELEXECUTIONREPLY"]._serialized_end = 2220
    _globals["_CANCANCELEXECUTIONREQUEST"]._serialized_start = 2222
    _globals["_CANCANCELEXECUTIONREQUEST"]._serialized_end = 2298
    _globals["_CANCANCELEXECUTIONREPLY"]._serialized_start = 2300
    _globals["_CANCANCELEXECUTIONREPLY"]._serialized_end = 2373
    _globals["_STARTRUNREQUEST"]._serialized_start = 2375
    _globals["_STARTRUNREQUEST"]._serialized_end = 2429
    _globals["_STARTRUNREPLY"]._serialized_start = 2431
    _globals["_STARTRUNREPLY"]._serialized_end = 2483
    _globals["_GETCURRENTIMAGEREPLY"]._serialized_start = 2485
    _globals["_GETCURRENTIMAGEREPLY"]._serialized_end = 2541
    _globals["_GETCURRENTRUNSREPLY"]._serialized_start = 2543
    _globals["_GETCURRENTRUNSREPLY"]._serialized_end = 2597
    _globals["_EXTERNALJOBREQUEST"]._serialized_start = 2599
    _globals["_EXTERNALJOBREQUEST"]._serialized_end = 2675
    _globals["_EXTERNALJOBREPLY"]._serialized_start = 2677
    _globals["_EXTERNALJOBREPLY"]._serialized_end = 2750
    _globals["_EXTERNALSCHEDULEEXECUTIONREPLY"]._serialized_start = 2752
    _globals["_EXTERNALSCHEDULEEXECUTIONREPLY"]._serialized_end = 2820
    _globals["_EXTERNALSENSOREXECUTIONREPLY"]._serialized_start = 2822
    _globals["_EXTERNALSENSOREXECUTIONREPLY"]._serialized_end = 2886
    _globals["_RELOADCODEREQUEST"]._serialized_start = 2888
    _globals["_RELOADCODEREQUEST"]._serialized_end = 2907
    _globals["_RELOADCODEREPLY"]._serialized_start = 2909
    _globals["_RELOADCODEREPLY"]._serialized_end = 2952
    _globals["_DAGSTERAPI"]._serialized_start = 2955
    _globals["_DAGSTERAPI"]._serialized_end = 5108
# @@protoc_insertion_point(module_scope)



# pyright: reportMissingImports=none
try:
    # zoneinfo is python >= 3.9
    from zoneinfo import ZoneInfo as _timezone_from_string
except:
    from dateutil.tz import gettz as _timezone_from_string
from datetime import tzinfo
from typing import Optional


def timezone_from_string(timezone_name: str) -> Optional[tzinfo]:
    # Allow case insensitivity for "utc" specifically for back-compat with pendulum 2
    # (plus the fact that some systems can process that timezone and others cannot)
    if timezone_name == "utc":
        timezone_name = "UTC"
    return _timezone_from_string(timezone_name)



from abc import ABC, abstractmethod
from functools import lru_cache
from typing import AbstractSet, Any, Mapping, Optional, Type

from typing_extensions import TypeVar

from dagster import _check as check
from dagster._model import DagsterModel
from dagster._model.pydantic_compat_layer import model_fields
from dagster._utils.typing_api import flatten_unions

from .metadata_value import MetadataValue, TableColumnLineage, TableSchema

T_NamespacedMetadataSet = TypeVar("T_NamespacedMetadataSet", bound="NamespacedMetadataSet")


# Python types that have a MetadataValue types that directly wraps them
DIRECTLY_WRAPPED_METADATA_TYPES = {
    str,
    float,
    int,
    bool,
    TableSchema,
    TableColumnLineage,
    type(None),
}

T_NamespacedMetadataSet = TypeVar("T_NamespacedMetadataSet", bound="NamespacedMetadataSet")


def is_raw_metadata_type(t: Type) -> bool:
    return issubclass(t, MetadataValue) or t in DIRECTLY_WRAPPED_METADATA_TYPES


class NamespacedMetadataSet(ABC, DagsterModel):
    """Extend this class to define a set of metadata fields in the same namespace.

    Supports splatting to a dictionary that can be placed inside a metadata argument along with
    other dictionary-structured metadata.

    .. code-block:: python

        my_metadata: NamespacedMetadataSet = ...
        return MaterializeResult(metadata={**my_metadata, ...})
    """

    def __init__(self, *args, **kwargs):
        for field_name in model_fields(self).keys():
            annotation_types = self._get_accepted_types_for_field(field_name)
            invalid_annotation_types = {
                annotation_type
                for annotation_type in annotation_types
                if not is_raw_metadata_type(annotation_type)
            }
            if invalid_annotation_types:
                check.failed(
                    f"Type annotation for field '{field_name}' includes invalid metadata type(s): {invalid_annotation_types}"
                )
        super().__init__(*args, **kwargs)

    @classmethod
    @abstractmethod
    def namespace(cls) -> str:
        raise NotImplementedError()

    @classmethod
    def _namespaced_key(cls, key: str) -> str:
        return f"{cls.namespace()}/{key}"

    @staticmethod
    def _strip_namespace_from_key(key: str) -> str:
        return key.split("/", 1)[1]

    def keys(self) -> AbstractSet[str]:
        return {
            self._namespaced_key(key)
            for key in model_fields(self).keys()
            # getattr returns the pydantic property on the subclass
            if getattr(self, key) is not None
        }

    def __getitem__(self, key: str) -> Any:
        # getattr returns the pydantic property on the subclass
        return getattr(self, self._strip_namespace_from_key(key))

    @classmethod
    def extract(
        cls: Type[T_NamespacedMetadataSet], metadata: Mapping[str, Any]
    ) -> T_NamespacedMetadataSet:
        """Extracts entries from the provided metadata dictionary into an instance of this class.

        Ignores any entries in the metadata dictionary whose keys don't correspond to fields on this
        class.

        In general, the following should always pass:

        .. code-block:: python

            class MyMetadataSet(NamedspacedMetadataSet):
                ...

            metadata: MyMetadataSet  = ...
            assert MyMetadataSet.extract(dict(metadata)) == metadata

        Args:
            metadata (Mapping[str, Any]): A dictionary of metadata entries.
        """
        kwargs = {}
        for namespaced_key, value in metadata.items():
            splits = namespaced_key.split("/")
            if len(splits) == 2:
                namespace, key = splits
                if namespace == cls.namespace() and key in model_fields(cls):
                    kwargs[key] = cls._extract_value(field_name=key, value=value)

        return cls(**kwargs)

    @classmethod
    def _extract_value(cls, field_name: str, value: Any) -> Any:
        """Based on type annotation, potentially coerce the metadata value to its inner value.

        E.g. if the annotation is Optional[float] and the value is FloatMetadataValue, construct
        the MetadataSet using the inner float.
        """
        if isinstance(value, MetadataValue):
            annotation = model_fields(cls)[field_name].annotation
            annotation_acceptable_types = flatten_unions(annotation)
            if (
                type(value) not in annotation_acceptable_types
                and type(value.value) in annotation_acceptable_types
            ):
                check.invariant(type(value.value) in DIRECTLY_WRAPPED_METADATA_TYPES)
                return value.value

        return value

    @classmethod
    @lru_cache(maxsize=None)  # this avoids wastefully recomputing this once per instance
    def _get_accepted_types_for_field(cls, field_name: str) -> AbstractSet[Type]:
        annotation = model_fields(cls)[field_name].annotation
        return flatten_unions(annotation)


class TableMetadataSet(NamespacedMetadataSet):
    """Metadata entries that apply to definitions, observations, or materializations of assets that
    are tables.

    Args:
        column_schema (Optional[TableSchema]): The schema of the columns in the table.
        column_lineage (Optional[TableColumnLineage]): The lineage of column inputs to column
            outputs for the table.
        row_count (Optional[int]): The number of rows in the table.
        partition_row_count (Optional[int]): The number of rows in the materialized or observed partition.
    """

    column_schema: Optional[TableSchema] = None
    column_lineage: Optional[TableColumnLineage] = None
    row_count: Optional[int] = None
    partition_row_count: Optional[int] = None

    @classmethod
    def namespace(cls) -> str:
        return "dagster"
